@@ -6,6 +6,7 @@
 #ifndef __MIME_H__
 #define __MIME_H__
 
+#include <config.h>
 #include <stddef.h>
 #include <gtk/gtk.h>
 #include "../dw_widget.h"
@@ -21,12 +22,18 @@ DwWidget* a_Html_text (const char *Type,void *web, CA_Callback_t *Call,
                        void **Data);
 DwWidget* a_Plain_text(const char *Type,void *web, CA_Callback_t *Call,
                        void **Data);
+#ifdef ENABLE_JPEG
 DwWidget* a_Jpeg_image(const char *Type,void *web, CA_Callback_t *Call,
                        void **Data);
+#endif
+#ifdef ENABLE_PNG
 DwWidget* a_Png_image (const char *Type,void *web, CA_Callback_t *Call,
                        void **Data);
+#endif
+#ifdef ENABLE_GIF
 DwWidget* a_Gif_image (const char *Type,void *web, CA_Callback_t *Call,
                        void **Data);
+#endif
 
 /*
  * Functions defined inside Mime module
