@@ -1,7 +1,7 @@
 /*
  * File: cache.c
  *
- * Copyright 2000, 2001 Jorge Arellano Cid <jcid@inf.utfsm.cl>
+ * Copyright 2000, 2001, 2002, 2003, 2004 Jorge Arellano Cid <jcid@dillo.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
 #include <fcntl.h>
 #include <unistd.h>
 
-#include "config.h"
+#include "msg.h"
 #include "list.h"
 #include "IO/Url.h"
 #include "IO/IO.h"
@@ -31,6 +31,7 @@
 #include "interface.h"
 #include "nav.h"
 #include "cookies.h"
+#include "misc.h"
 
 #define NULLKey 0
 
@@ -67,6 +68,11 @@ static GHashTable *CacheHash = NULL;
  * Although implemented as a list, we'll call it ClientQueue  --Jcid */
 static GSList *ClientQueue = NULL;
 
+/* A list for delayed clients (it holds weak pointers to cache entries,
+ * which are used to make deferred calls to Cache_process_queue) */
+static GSList *DelayedQueue = NULL;
+static guint DelayedQueueIdleId = 0;
+
 
 /*
  *  Forward declarations
@@ -210,6 +216,15 @@ static void Cache_entry_init(CacheData_t *NewEntry, const DilloUrl *Url)
    NewEntry->CCCAnswer = NULL;
 }
 
+/*
+ * Get the data structure for a cached URL (using 'Url' as the search key)
+ * If 'Url' isn't cached, return NULL
+ */
+static CacheData_t *Cache_entry_search(const DilloUrl *Url)
+{
+   return g_hash_table_lookup(CacheHash, Url);
+}
+
 /*
  * Allocate and set a new entry in the cache list
  */
@@ -217,21 +232,14 @@ static CacheData_t *Cache_entry_add(const DilloUrl *Url)
 {
    CacheData_t *new_entry = g_new(CacheData_t, 1);
 
+   if (Cache_entry_search(Url))
+      DEBUG_MSG(5, "WARNING: Cache_entry_add, leaking an entry.\n");
+
    Cache_entry_init(new_entry, Url);  /* Set safe values */
    g_hash_table_insert(CacheHash, (gpointer)new_entry->Url, new_entry);
    return new_entry;
 }
 
-/*
- * Get the data structure for a cached URL (using 'Url' as the search key)
- * If 'Url' isn't cached, return NULL
- * (We search backwards to speed up the process)
- */
-static CacheData_t *Cache_entry_search(const DilloUrl *Url)
-{
-   return g_hash_table_lookup(CacheHash, Url);
-}
-
 /*
  *  Free the components of a CacheData_t struct.
  */
@@ -258,6 +266,9 @@ static gint Cache_entry_remove_raw(CacheData_t *entry, DilloUrl *url)
    g_return_val_if_fail(
       !g_slist_find_custom(ClientQueue, url, Cache_client_url_cmp), 0);
 
+   /* remove from DelayedQueue */
+   DelayedQueue = g_slist_remove(DelayedQueue, entry);
+
    /* remove from dicache */
    a_Dicache_invalidate_entry(url);
 
@@ -283,10 +294,10 @@ static void Cache_entry_remove(CacheData_t *entry, DilloUrl *url)
 
    if (InfoQuery) {
       DEBUG_MSG(2, "## Aborting CCCQuery\n");
-      a_Cache_ccc(OpAbort, -1, InfoQuery, NULL, NULL);
+      a_Cache_ccc(OpAbort, 1, BCK, InfoQuery, NULL, NULL);
    } else if (InfoAnswer) {
       DEBUG_MSG(2, "## Aborting CCCAnswer\n");
-      a_Cache_ccc(OpAbort, -1, InfoAnswer, NULL, NULL);
+      a_Cache_ccc(OpAbort, 2, BCK, InfoAnswer, NULL, NULL);
    } else {
       DEBUG_MSG(2, "## Aborting raw2\n");
       Cache_entry_remove_raw(entry, url);
@@ -335,13 +346,12 @@ static gint Cache_prepare_reload(DilloUrl *url)
       Cache_entry_remove(NULL, url);
       return 1;
    } else {
-      g_print("Cache_prepare_reload: ERROR, entry still has clients\n");
+      MSG("Cache_prepare_reload: ERROR, entry still has clients\n");
    }
 
    return 0;
 }
 
-
 /*
  * Try finding the url in the cache. If it hits, send the cache contents
  * from there. If it misses, set up a new connection.
@@ -355,25 +365,29 @@ static gint Cache_open_url(DilloWeb *Web, CA_Callback_t Call, void *CbData)
    DilloUrl *Url = Web->url;
    CacheData_t *entry = Cache_entry_search(Url);
 
-   if ( !entry ) {
+   _MSG("Cache_open_url: %s %sFOUND\n", URL_STR(Url), entry ? "" : "NOT ");
+
+   if ( entry ) {
+      /* URL is cached: feed our client with cached data */
+      ClientKey = Cache_client_enqueue(entry->Url, Web, Call, CbData);
+      Cache_delayed_process_queue(entry);
+
+   } else {
       /* URL not cached: create an entry, send our client to the queue,
        * and open a new connection */
       entry = Cache_entry_add(Url);
       ClientKey = Cache_client_enqueue(entry->Url, Web, Call, CbData);
-      a_Cache_ccc(OpStart, 1, entry->CCCQuery, NULL, (void *)entry->Url);
+      a_Cache_ccc(OpStart, 1, BCK, entry->CCCQuery, NULL, (void *)entry->Url);
       cccFunct = a_Url_get_ccc_funct(entry->Url);
       if ( cccFunct ) {
-         link = a_Chain_link_new(a_Cache_ccc, entry->CCCQuery,
-                                 CCC_FWD, cccFunct);
-         cccFunct(OpStart, 1, link, (void *)entry->Url, Web);
+         link = a_Chain_link_new(entry->CCCQuery,
+                                 a_Cache_ccc, BCK, cccFunct, 1, 1);
+         a_Chain_bcb(OpStart, entry->CCCQuery, (void *)entry->Url, Web);
       } else {
          a_Interface_msg(Web->bw, "ERROR: unsupported protocol");
-         a_Cache_ccc(OpAbort, 1, entry->CCCQuery, NULL, NULL);
+         a_Cache_ccc(OpAbort, 1, FWD, entry->CCCQuery, NULL, NULL);
+         ClientKey = 0; /* aborted */
       }
-   } else {
-      /* Feed our client with cached data */
-      ClientKey = Cache_client_enqueue(entry->Url, Web, Call, CbData);
-      Cache_delayed_process_queue(entry);
    }
    return ClientKey;
 }
@@ -427,15 +441,15 @@ gint a_Cache_open_url(void *web, CA_Callback_t Call, void *CbData)
 }
 
 /*
- * Return the pointer to URL 'Data' in the cache.
- * (We also return its size in 'Size' and take care of redirections)
+ * Get the pointer to the URL document, and its size, from the cache entry.
+ * Return: 1 cached, 0 not cached.
  */
-char *a_Cache_url_read(const DilloUrl *Url, gint *Size)
+gint a_Cache_get_buf(const DilloUrl *Url, gchar **PBuf, gint *BufSize)
 {
    CacheData_t **stack = NULL;
    gint stackSize = 0;
    gint stackMax = 32;
-   CacheData_t *entry = NULL;
+   CacheData_t *entry;
    gint i;
 
    do {
@@ -457,10 +471,11 @@ char *a_Cache_url_read(const DilloUrl *Url, gint *Size)
          }
       }
    } while (entry);
-
    g_free (stack);
-   *Size = (entry) ? entry->ValidSize : 0;
-   return (entry) ? (char *) entry->Data : NULL;
+
+   *BufSize = (entry) ? entry->ValidSize : 0;
+   *PBuf = (entry) ? (gchar *) entry->Data : NULL;
+   return (entry ? 1 : 0);
 }
 
 /*
@@ -501,7 +516,8 @@ static char *Cache_parse_field(const char *header, const char *fieldname)
 /*
  * Extract multiple fields from the header.
  */
-GList *Cache_parse_multiple_fields(const char *header, const char *fieldname)
+static GList *Cache_parse_multiple_fields(const char *header,
+                                          const char *fieldname)
 {
    gint i, j;
    GList *fields = NULL;
@@ -531,55 +547,6 @@ GList *Cache_parse_multiple_fields(const char *header, const char *fieldname)
    return fields;
 }
 
-/*
- * Detects 'Content-Type' when the server does not supply one.
- * It uses the magic(5) logic from file(1). Currently, it
- * only checks the few mime types that Dillo supports.
- *
- * 'Data' is a pointer to the first bytes of the raw data.
- */
-static gchar *Cache_get_type_from_data(void *Data, size_t Size)
-{
-   static gchar *Types[] = {
-      "application/octet-stream",
-      "text/html", "text/plain",
-      "image/gif", "image/png", "image/jpeg",
-   };
-   gint Type = 0;
-   gchar *p = Data;
-   int i;
-
-   /* HTML try */
-   for (i = 0; i < Size && isspace(p[i]); ++i);
-   if ((Size - i >= 5 && !g_strncasecmp(p+i, "<html", 5)) ||
-       (Size - i >= 5 && !g_strncasecmp(p+i, "<head", 5)) ||
-       (Size - i >= 6 && !g_strncasecmp(p+i, "<title", 6)) ||
-       (Size - i >= 14 && !g_strncasecmp(p+i, "<!doctype html", 14))) {
-      Type = 1;
-
-   /* Images */
-   } else if (Size >= 4 && !g_strncasecmp(p, "GIF8", 4)) {
-      Type = 3;
-   } else if (Size >= 4 && !g_strncasecmp(p, "\x89PNG", 4)) {
-      Type = 4;
-   } else if (Size >= 2 && !g_strncasecmp(p, "\xff\xd8", 2)) {
-      /* JPEG has the first 2 bytes set to 0xffd8 in BigEndian - looking
-       * at the character representation should be machine independent. */
-      Type = 5;
-
-   /* Text */
-   } else {
-      /* We'll assume ASCII if chars are below 128 (after all, this
-       * is a last resort when the server doesn't send Content-Type) */
-      for (i = 0; i < Size && i < 256; i++)
-         if ((guchar) p[i] > 127)
-            break;
-      Type = (i < 12 || (guchar) p[i] > 127) ? 0 : 2;
-   }
-
-   return g_strdup(Types[Type]);
-}
-
 /*
  * Scan, allocate, and set things according to header info.
  * (This function needs the whole header to work)
@@ -588,16 +555,21 @@ static void Cache_parse_header(CacheData_t *entry, IOData_t *io, gint HdrLen)
 {
    gchar *header = entry->Header->str;
    gchar *Length, *Type, *location_str;
+#ifndef DISABLE_COOKIES
    GList *Cookies;
+#endif
 
-   if ( header[9] == '3' && header[10] == '0' ) {
+   if ( HdrLen < 12 ) {
+      /* Not enough info. */
+
+   } if ( header[9] == '3' && header[10] == '0' ) {
       /* 30x: URL redirection */
       entry->Flags |= CA_Redirect;
       if ( header[11] == '1' )
          /* 301 Moved Permanently */
          entry->Flags |= CA_ForceRedirect;
       location_str = Cache_parse_field(header, "Location");
-      entry->Location = a_Url_new(location_str, URL_STR_(entry->Url), 0, 0);
+      entry->Location = a_Url_new(location_str, URL_STR_(entry->Url), 0, 0, 0);
       g_free(location_str);
 
    } else if ( strncmp(header + 9, "404", 3) == 0 ) {
@@ -627,30 +599,24 @@ static void Cache_parse_header(CacheData_t *entry, IOData_t *io, gint HdrLen)
 
    if ( entry->TotalSize > 0 && entry->TotalSize >= entry->ValidSize ) {
       entry->Data = g_malloc(entry->TotalSize);
-      memcpy(entry->Data, (char*)io->IOVec.iov_base+HdrLen, io->Status-HdrLen);
-      /* Free preallocated buffer */
-      if (io->Flags & IOFlag_FreeIOVec)
-         g_free(io->IOVec.iov_base);
+      memcpy(entry->Data, (char*)io->Buf+HdrLen, io->Status-HdrLen);
       /* Prepare next read */
-      io->IOVec.iov_base = (char *)entry->Data + entry->ValidSize;
-      io->IOVec.iov_len  = entry->TotalSize - entry->ValidSize;
-      io->Flags &= ~IOFlag_FreeIOVec;
+      a_IO_set_buf(io, (char *)entry->Data + entry->ValidSize,
+                   entry->TotalSize - entry->ValidSize);
    } else {
       /* We don't know the size of the transfer; A lazy server? ;) */
       entry->Data = g_malloc(entry->ValidSize + entry->BuffSize);
-      memcpy(entry->Data, (char *)io->IOVec.iov_base+HdrLen, entry->ValidSize);
+      memcpy(entry->Data, (char *)io->Buf+HdrLen, entry->ValidSize);
       /* Prepare next read */
-      if (io->Flags & IOFlag_FreeIOVec)
-         g_free(io->IOVec.iov_base);
-      io->IOVec.iov_base = (char *)entry->Data + entry->ValidSize;
-      io->IOVec.iov_len  = entry->BuffSize;
-      io->Flags &= ~IOFlag_FreeIOVec;
+      a_IO_set_buf(io, (char *)entry->Data + entry->ValidSize,
+                   entry->BuffSize);
    }
 
    /* Get Content-Type */
    if ( (Type = Cache_parse_field(header, "Content-Type")) == NULL ) {
-     DEBUG_HTTP_MSG("Server didn't send Content-Type in header.\n");
-     Type = Cache_get_type_from_data(entry->Data, entry->ValidSize);
+     MSG_HTTP("Server didn't send Content-Type in header.\n");
+     Type = g_strdup(a_Misc_get_content_type_from_data(
+                        entry->Data, entry->ValidSize));
    }
    entry->Type = Type;
 }
@@ -663,7 +629,7 @@ static gint Cache_get_header(IOData_t *io, CacheData_t *entry)
 {
    gint N, i;
    GString *hdr = entry->Header;
-   guchar *data = io->IOVec.iov_base;
+   guchar *data = io->Buf;
 
    /* Header finishes when N = 2 */
    N = (hdr->len && hdr->str[hdr->len - 1] == '\n');
@@ -708,8 +674,8 @@ static void Cache_process_io(int Op, void *VPtr)
 
    if ( Op == IOClose ) {
       if (entry->Flags & CA_GotLength && entry->TotalSize != entry->ValidSize){
-         DEBUG_HTTP_MSG("Content-Length does NOT match message body,\n"
-                        " at: %s\n", URL_STR_(entry->Url));
+         MSG_HTTP("Content-Length does NOT match message body,\n"
+                  " at: %s\n", URL_STR_(entry->Url));
       }
       entry->Flags |= CA_GotData;
       entry->Flags &= ~CA_Stopped;          /* it may catch up! */
@@ -741,21 +707,20 @@ static void Cache_process_io(int Op, void *VPtr)
 
    Status = io->Status;
    entry->ValidSize += Status;
-   if ( Status < (gint)io->IOVec.iov_len ) {
+   if ( Status < (gint)io->BufSize ) {
       /* An incomplete buffer; update buffer & size */
-      io->IOVec.iov_len  -= Status;
-      io->IOVec.iov_base = (char *)io->IOVec.iov_base + Status;
-   } else if ( Status == (gint)io->IOVec.iov_len ) {
+      a_IO_set_buf(io, (char *)io->Buf + Status, io->BufSize - Status);
+
+   } else if ( Status == (gint)io->BufSize ) {
       /* A full buffer! */
       if ( !entry->TotalSize ) {
          /* We are receiving in small chunks... */
          entry->Data = g_realloc(entry->Data,entry->ValidSize+entry->BuffSize);
-         io->IOVec.iov_base = (char *)entry->Data + entry->ValidSize;
-         io->IOVec.iov_len  = entry->BuffSize;
+         a_IO_set_buf(io, (char *)entry->Data + entry->ValidSize,
+                      entry->BuffSize);
       } else {
          /* We have a preallocated buffer! */
-         io->IOVec.iov_len  -= Status;
-         io->IOVec.iov_base = (char *)io->IOVec.iov_base + Status;
+         a_IO_set_buf(io, (char *)io->Buf + Status, io->BufSize - Status);
       }
    }
    Cache_process_queue(entry);
@@ -786,7 +751,8 @@ static gint Cache_redirect(CacheData_t *entry, gint Flags, BrowserWindow *bw)
       if ( Flags & WEB_RootUrl ) {
          /* Redirection of the main page */
          a_Nav_remove_top_url(bw);
-         NewUrl =a_Url_new(URL_STR_(entry->Location),URL_STR_(entry->Url),0,0);
+         NewUrl = a_Url_new(URL_STR_(entry->Location), URL_STR_(entry->Url),
+                            0, 0, 0);
          a_Nav_push(bw, NewUrl);
          a_Url_free(NewUrl);
       } else {
@@ -822,12 +788,12 @@ void a_Cache_null_client(int Op, CacheClient_t *Client)
  */
 static void Cache_process_queue(CacheData_t *entry)
 {
-   gint i;
+   gint i, st;
    CacheClient_t *Client;
    DilloWeb *ClientWeb;
+   BrowserWindow *Client_bw = NULL;
    static gboolean Busy = FALSE;
-   gboolean EntryHasClients = FALSE;
-   const DilloUrl *Url = entry->Url;
+   gboolean AbortEntry = FALSE;
 
    if ( Busy )
       DEBUG_MSG(5, "FATAL!:*** >>>> Cache_process_queue Caught busy!!!\n");
@@ -838,24 +804,36 @@ static void Cache_process_queue(CacheData_t *entry)
    }
 
    for ( i = 0; (Client = g_slist_nth_data(ClientQueue, i)); ++i ) {
-      if ( Client->Url == Url ) {
-         EntryHasClients = TRUE;
-         ClientWeb = Client->Web; /* It was a (void*) */
+      if ( Client->Url == entry->Url ) {
+         ClientWeb = Client->Web;    /* It was a (void*) */
+         Client_bw = ClientWeb->bw;  /* 'bw' in a local var */
 
-         /* For root URLs, clear the "expecting for reply..." message */
-         if (ClientWeb->flags & WEB_RootUrl && !(entry->Flags & CA_MsgErased)){
-            a_Interface_msg(ClientWeb->bw, "");
-            entry->Flags |= CA_MsgErased;
+         if (ClientWeb->flags & WEB_RootUrl) {
+            if (!(entry->Flags & CA_MsgErased)) {
+               /* clear the "expecting for reply..." message */
+               a_Interface_msg(Client_bw, "");
+               entry->Flags |= CA_MsgErased;
+            }
+         } else {
+            /* For non root URLs, ignore redirections and 404 answers */
+            if (entry->Flags & CA_Redirect || entry->Flags & CA_NotFound)
+               Client->Callback = a_Cache_null_client;
+         }
+
+         /* Set the client function */
+         if (!Client->Callback) {
+            st = a_Web_dispatch_by_type(entry->Type, ClientWeb,
+                                        &Client->Callback, &Client->CbData);
+            if (st == -1 && ClientWeb->flags & WEB_RootUrl) {
+               /* Unhandled MIME type, prepare a download offer... */
+               AbortEntry = TRUE;
+               a_Interface_remove_client(Client_bw, Client->Key);
+               Cache_client_dequeue(Client, NULLKey);
+               --i; /* Keep the index value in the next iteration */
+               continue;
+            }
          }
-         /* For non root URLs, ignore redirections and 404 answers */
-         if ( !(ClientWeb->flags & WEB_RootUrl) &&
-              (entry->Flags & CA_Redirect || entry->Flags & CA_NotFound) )
-            Client->Callback = a_Cache_null_client;
 
-         /* Set client function */
-         if ( !Client->Callback )
-            a_Web_dispatch_by_type(entry->Type, ClientWeb,
-                                   &Client->Callback, &Client->CbData);
          /* Send data to our client */
          if ( (Client->BufSize = entry->ValidSize) > 0) {
             Client->Buf = (guchar *)entry->Data;
@@ -864,8 +842,7 @@ static void Cache_process_queue(CacheData_t *entry)
 
          /* Remove client when done */
          if ( (entry->Flags & CA_GotData) ) {
-            /* Copy Client data to local vars */
-            void *bw = ClientWeb->bw;
+            /* Copy flags to a local var */
             gint flags = ClientWeb->flags;
             /* We finished sending data, let the client know */
             if (!Client->Callback)
@@ -875,11 +852,18 @@ static void Cache_process_queue(CacheData_t *entry)
             Cache_client_dequeue(Client, NULLKey);
             --i; /* Keep the index value in the next iteration */
             if ( entry->Flags & CA_Redirect )
-               Cache_redirect(entry, flags, bw);
+               Cache_redirect(entry, flags, Client_bw);
          }
       }
    } /* for */
 
+   if (AbortEntry) {
+      /* Abort the entry, remove it from cache, and offer download.
+       * (the dialog is made before 'entry' is freed) */
+      a_Interface_offer_link_download(Client_bw, entry->Url);
+      Cache_entry_remove(entry, NULL);
+   }
+
    Busy = FALSE;
    DEBUG_MSG(1, "QueueSize ====> %d\n", g_slist_length(ClientQueue));
 }
@@ -889,16 +873,30 @@ static void Cache_process_queue(CacheData_t *entry)
  */
 static gint Cache_delayed_process_queue_callback(gpointer data)
 {
-   Cache_process_queue( (CacheData_t *)data );
+   gpointer entry;
+
+   while ((entry = g_slist_nth_data(DelayedQueue, 0))) {
+      Cache_process_queue( (CacheData_t *)entry );
+      /* note that if Cache_process_queue removes the entry,
+       * the following g_slist_remove has no effect. */
+      DelayedQueue = g_slist_remove(DelayedQueue, entry);
+   }
+   DelayedQueueIdleId = 0;
    return FALSE;
 }
 
 /*
- * Call Cache_process_queue from the gtk_main cycle
+ * Set a call to Cache_process_queue from the gtk_main cycle.
  */
 static void Cache_delayed_process_queue(CacheData_t *entry)
 {
-   gtk_idle_add((GtkFunction)Cache_delayed_process_queue_callback, entry);
+   /* there's no need to repeat entries in the queue */
+   if (!g_slist_find(DelayedQueue, entry))
+      DelayedQueue = g_slist_prepend(DelayedQueue, entry);
+
+   if (DelayedQueueIdleId == 0)
+      DelayedQueueIdleId =
+         gtk_idle_add((GtkFunction)Cache_delayed_process_queue_callback, NULL);
 }
 
 
@@ -956,9 +954,9 @@ static void Cache_stop_client(gint Key, gint force)
       /* it was the last client of this entry */
       if ((entry = Cache_entry_search(url))) {
          if (entry->CCCQuery) {
-            a_Cache_ccc(OpAbort, -1, entry->CCCQuery, NULL, NULL);
+            a_Cache_ccc(OpAbort, 1, BCK, entry->CCCQuery, NULL, NULL);
          } else if (entry->CCCAnswer) {
-            a_Cache_ccc(OpStop, -1, entry->CCCAnswer, NULL, Client);
+            a_Cache_ccc(OpStop, 2, BCK, entry->CCCAnswer, NULL, Client);
          }
       }
    }
@@ -977,94 +975,122 @@ void a_Cache_stop_client(gint Key)
 /*
  * CCC function for the CACHE module
  */
-void
- a_Cache_ccc(int Op, int Branch, ChainLink *Info, void *Data, void *ExtraData)
+void a_Cache_ccc(int Op, int Branch, int Dir, ChainLink *Info,
+                 void *Data1, void *Data2)
 {
    CacheData_t *entry;
 
+   a_Chain_debug_msg("a_Cache_ccc", Op, Branch, Dir);
+
    if ( Branch == 1 ) {
-      /* Querying branch */
-      switch (Op) {
-      case OpStart:
-         DEBUG_MSG(3, "Cache CCC (OpStart) [0]");
-         Info->LocalKey = ExtraData;
-         break;
-      case OpEnd:
-         /* unlink HTTP_Info */
-         a_Chain_del_link(Info, CCC_BCK);
-         /* 'entry->CCCQuery' and 'Info' point to the same place! */
-         if ((entry = Cache_entry_search(Info->LocalKey)) != NULL)
-            entry->CCCQuery = NULL;
-         g_free(Info);
-         DEBUG_MSG(3, "Cache CCC (OpEnd) [0]\n");
-         break;
-      case OpAbort:
-         /* Unlink HTTP_Info */
-         a_Chain_del_link(Info, CCC_BCK);
+      if (Dir == BCK) {
+         /* Querying branch */
+         switch (Op) {
+         case OpStart:
+            /* Localkey = entry->Url */
+            Info->LocalKey = Data2;
+            break;
+         case OpStop:
+            break;
+         case OpAbort:
+            Cache_entry_remove_raw(NULL, Info->LocalKey);
+            a_Chain_bcb(OpAbort, Info, NULL, NULL);
+            g_free(Info);
+            break;
+         }
+      } else {  /* FWD */
+         switch (Op) {
+         case OpSend:
+            /* Start the answer-reading chain */
+            a_Cache_ccc(OpStart, 2, BCK, a_Chain_new(), Data1, Info->LocalKey);
+            break;
+         case OpEnd:
+            /* unlink HTTP_Info */
+            a_Chain_del_link(Info, BCK);
+            /* 'entry->CCCQuery' and 'Info' point to the same place! */
+            if ((entry = Cache_entry_search(Info->LocalKey)) != NULL)
+               entry->CCCQuery = NULL;
+            g_free(Info);
+            break;
+         case OpAbort:
+            /* Unlink HTTP_Info */
+            a_Chain_del_link(Info, BCK);
 
-         /* remove interface client-references of this entry */
-         Cache_remove_interface_clients(Info->LocalKey);
-         /* remove clients of this entry */
-         Cache_stop_clients(Info->LocalKey, 0);
-         /* remove the entry */
-         Cache_entry_remove_raw(NULL, Info->LocalKey);
+            /* remove interface client-references of this entry */
+            Cache_remove_interface_clients(Info->LocalKey);
+            /* remove clients of this entry */
+            Cache_stop_clients(Info->LocalKey, 0);
+            /* remove the entry */
+            Cache_entry_remove_raw(NULL, Info->LocalKey);
 
-         g_free(Info);
-         DEBUG_MSG(3, "Cache CCC (OpAbort) [0]\n");
-         break;
+            g_free(Info);
+            break;
+         }
       }
 
    } else if (Branch == 2) {
-      /* Answering branch */
-      switch (Op) {
-      case OpStart:
-         Info->LocalKey = ExtraData;
-         if ((entry = Cache_entry_search(Info->LocalKey))) {
-            entry->CCCAnswer = Info;
-         } else {
-            /* The cache-entry was removed */
-            a_Chain_bcb(OpAbort, -1, Info, NULL, NULL);
+      if (Dir == FWD) {
+         /* Answering branch */
+         switch (Op) {
+         case OpStart:
+            /* Data2 = entry->Url */
+            Info->LocalKey = Data2;
+            if ((entry = Cache_entry_search(Info->LocalKey))) {
+               entry->CCCAnswer = Info;
+            } else {
+               /* The cache-entry was removed */
+               a_Chain_bcb(OpAbort, Info, NULL, NULL);
+               g_free(Info);
+            }
+            break;
+         case OpSend:
+            /* Send data */
+            if ((entry = Cache_entry_search(Info->LocalKey))) {
+               Cache_process_io(IORead, Data1);
+            } else {
+               a_Chain_bcb(OpAbort, Info, NULL, NULL);
+               g_free(Info);
+            }
+            break;
+         case OpEnd:
+            /* Unlink HTTP_Info */
+            a_Chain_del_link(Info, BCK);
             g_free(Info);
-         }
-         break;
-      case OpSend:
-         /* Send data */
-         if ((entry = Cache_entry_search(Info->LocalKey))) {
-            Cache_process_io(IORead, Data);
-         } else {
-            a_Chain_bcb(OpAbort, -1, Info, NULL, NULL);
+            Cache_process_io(IOClose, Data1);
+            break;
+         case OpAbort:
+            a_Chain_del_link(Info, BCK);
             g_free(Info);
+            Cache_process_io(IOAbort, Data1);
+            break;
          }
-         break;
-      case OpEnd:
-         /* Unlink HTTP_Info */
-         a_Chain_del_link(Info, CCC_BCK);
-         g_free(Info);
-         Cache_process_io(IOClose, Data);
-         DEBUG_MSG(3, "Cache CCC (OpEnd) [1]\n");
-         break;
-      case OpAbort:
-         a_Chain_del_link(Info, CCC_BCK);
-         g_free(Info);
-         Cache_process_io(IOAbort, Data);
-         DEBUG_MSG(3, "Cache CCC (OpAbort) [1]\n");
-         break;
-      }
+      } else {  /* BCK */
+         switch (Op) {
+         case OpStart:
+            {
+            IOData_t *io2;
 
-   } else if (Branch == -1) {
-      /* Backwards operations */
-      switch (Op) {
-      case OpStop:
-         /* We'll let it fill the entry by now... */
-         /* a_Chain_bcb(OpStop, -1, Info, NULL, NULL);
-            g_free(Info); */
-         break;
-      case OpAbort:
-         DEBUG_MSG(2, "Cache: OpAbort [-1]\n");
-         Cache_entry_remove_raw(NULL, Info->LocalKey);
-         a_Chain_bcb(OpAbort, -1, Info, NULL, NULL);
-         g_free(Info);
-         break;
+            Info->LocalKey = Data2;
+            if ((entry = Cache_entry_search(Data2)))    // Is Data2 valid?
+               entry->CCCAnswer = Info;
+
+            /* Link IO to receive the answer */
+            io2 = a_IO_new(IORead, *(int*)Data1);
+            a_IO_set_buf(io2, NULL, IOBufLen);
+            io2->ExtData = Data2;       // We have it as LocalKey
+            a_Chain_link_new(Info, a_Cache_ccc, BCK, a_IO_ccc, 2, 2);
+            a_Chain_bcb(OpStart, Info, io2, NULL);
+            break;
+            }
+         case OpStop:
+            MSG(" Not implemented\n");
+            break;
+         case OpAbort:
+            Cache_entry_remove_raw(NULL, Info->LocalKey);
+            a_Chain_bcb(OpAbort, Info, NULL, NULL);
+            g_free(Info);
+            break;
+         }
       }
    }
 }
