@@ -34,13 +34,17 @@ struct _DwAlignedPageClass
 
    gint32 (*get_value)     (DwAlignedPage *aligned_page);
    void   (*set_max_value) (DwAlignedPage *aligned_page,
+                            gint32 max_value,
                             gint32 value);
 };
 
 
 struct _DwAlignedPageList
 {
-   gint num;
+   gint num;      /* The index of the last page added (minus one). */
+   gint refcount; /* The numbers of non-NULL pages, may be smaller than num.
+                   * The distinction is necessary, because removing is
+                   * simply done by assigning NULL. */
    DwAlignedPage **pages;
    gint32 *values;
    gint num_pages_max;
