@@ -1,6 +1,7 @@
 #include <gtk/gtk.h>
 #include "dw_embed_gtk.h"
 #include "dw_gtk_viewport.h"
+#include "debug.h"
 
 
 static void Dw_embed_gtk_init           (DwEmbedGtk *embed_gtk);
@@ -55,6 +56,7 @@ DwWidget* a_Dw_embed_gtk_new (void)
    GtkObject *object;
 
    object = gtk_object_new (DW_TYPE_EMBED_GTK, NULL);
+   DBG_OBJ_CREATE (object, "DwEmbedGtk");
    return DW_WIDGET (object);
 }
 
@@ -65,7 +67,8 @@ DwWidget* a_Dw_embed_gtk_new (void)
 static void Dw_embed_gtk_init (DwEmbedGtk *embed_gtk)
 {
    embed_gtk->child = NULL;
-   embed_gtk->idle_id = -1;
+   embed_gtk->idle_id = 0;
+   /*embed_gtk->child_size_request_lock = FALSE;*/
 }
 
 
@@ -99,10 +102,10 @@ static void Dw_embed_gtk_destroy (GtkObject *object)
    embed_gtk = DW_EMBED_GTK (object);
    if (embed_gtk->child)
       gtk_object_destroy (GTK_OBJECT (embed_gtk->child));
-   if (embed_gtk->idle_id != -1)
+   if (embed_gtk->idle_id != 0)
       gtk_idle_remove (embed_gtk->idle_id);
 
-   (* GTK_OBJECT_CLASS(parent_class)->destroy) (object);
+   GTK_OBJECT_CLASS(parent_class)->destroy (object);
 }
 
 
@@ -118,7 +121,9 @@ static void Dw_embed_gtk_size_request   (DwWidget *widget,
 
    embed_gtk = DW_EMBED_GTK (widget);
    if (embed_gtk->child) {
+      /*embed_gtk->child_size_request_lock = TRUE;*/
       gtk_widget_size_request (embed_gtk->child, &child_requisition);
+      /*embed_gtk->child_size_request_lock = FALSE;*/
 
       requisition->width = child_requisition.width;
       requisition->ascent = child_requisition.height;
@@ -151,7 +156,7 @@ static gint Dw_embed_gtk_move_idle (void *data)
                          widget->allocation.x, widget->allocation.y);
    }
 
-   embed_gtk->idle_id = -1;
+   embed_gtk->idle_id = 0;
    return FALSE;
 }
 
@@ -175,7 +180,7 @@ static void Dw_embed_gtk_size_allocate (DwWidget *widget,
 
          if (allocation->width != widget->allocation.width ||
              allocation->ascent + allocation->descent
-             != widget->allocation.ascent + widget->allocation.descent)
+             != DW_WIDGET_HEIGHT(widget))
             /* todo: try implementing it by gtk_widget_size_allocate */
             gtk_widget_set_usize (embed_gtk->child,
                                   allocation->width,
@@ -185,7 +190,7 @@ static void Dw_embed_gtk_size_allocate (DwWidget *widget,
              allocation->y != widget->allocation.y) {
             /* A simple call to gtk_layout_move does not seem to work, so this
                strange idle function. */
-            if (embed_gtk->idle_id == -1)
+            if (embed_gtk->idle_id == 0)
                embed_gtk->idle_id =
                   gtk_idle_add (Dw_embed_gtk_move_idle, (void*) embed_gtk);
          }
@@ -235,7 +240,7 @@ static int Dw_embed_gtk_child_focus_in (DwEmbedGtk *embed_gtk)
    wx = widget->allocation.x;
    wy = widget->allocation.y;
    ww = widget->allocation.width;
-   wh = widget->allocation.ascent + widget->allocation.descent;
+   wh = DW_WIDGET_HEIGHT(widget);
 
    if (vx > wx)
       gtk_adjustment_set_value (layout->hadjustment, wx);
@@ -250,6 +255,33 @@ static int Dw_embed_gtk_child_focus_in (DwEmbedGtk *embed_gtk)
    return FALSE;
 }
 
+#if 0
+/*
+ * Called when the embedded Gtk widget emits a "size_request" signal, and used
+ * to resise the embedding Dw widget.
+ *
+ * Currently not used.
+ */
+static void Dw_embed_gtk_child_size_request (DwEmbedGtk *embed_gtk)
+{
+   DwWidget *widget;
+   GtkRequisition child_requisition;
+
+   g_return_if_fail (embed_gtk->child);
+
+   if(!embed_gtk->child_size_request_lock) {
+      widget = DW_WIDGET (embed_gtk);
+      embed_gtk->child_size_request_lock = TRUE;
+      gtk_widget_size_request (embed_gtk->child, &child_requisition);
+      embed_gtk->child_size_request_lock = FALSE;
+      
+      /* Check wether the size has changed at all. */
+      if (widget->requisition.width != child_requisition.width ||
+          widget->requisition.ascent != child_requisition.height)
+         p_Dw_widget_queue_resize (widget, 0, TRUE);
+   }
+}
+#endif
 
 /*
  * Add the Gtk+ widget to be embedded.
@@ -270,13 +302,13 @@ void a_Dw_embed_gtk_add_gtk (DwEmbedGtk *embed_gtk,
                       DW_WIDGET (embed_gtk)->allocation.x,
                       DW_WIDGET (embed_gtk)->allocation.y);
 
+   /* todo: We somehow must regognize size changes of the embedded Gtk+
+    * widget. Can this be done with size_request? Anyway, currently
+    * deactivated, since it is not used in dillo. */
 #if 0
-   /* This is to recognize size changes of the embedded Gtk+ widget,
-      but this simple implementation causes too much page rewraps, so
-      it is currently deactivated */
    gtk_signal_connect_object (GTK_OBJECT (embed_gtk->child),
                               "size_request",
-                              GTK_SIGNAL_FUNC (Dw_widget_queue_resize),
+                              GTK_SIGNAL_FUNC(Dw_embed_gtk_child_size_request),
                               GTK_OBJECT (embed_gtk));
 #endif
 
