@@ -1,7 +1,7 @@
 /*
  * File: dw_gtk_viewport.c
  *
- * Copyright (C) 2001  Sebastian Geerken <S.Geerken@ping.de>
+ * Copyright (C) 2001-2003  Sebastian Geerken <S.Geerken@ping.de>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -10,10 +10,20 @@
  */
 
 #include <gtk/gtk.h>
+#include "msg.h"
 #include "dw_gtk_viewport.h"
 #include "dw_container.h"
+#include "list.h"
 
+/*#define DEBUG_LEVEL 1*/
 #include "debug.h"
+#include "msg.h"
+
+typedef struct
+{
+   DwWidget *widget;
+   gint32 y;
+} GtkDwViewportAnchor;
 
 static GtkLayoutClass *parent_class = NULL;
 
@@ -108,6 +118,8 @@ GtkWidget* a_Dw_gtk_viewport_new (GtkAdjustment *hadjustment,
  */
 static void Dw_gtk_viewport_init (GtkDwViewport *viewport)
 {
+   DBG_OBJ_CREATE (viewport, "GtkDwViewport");
+
    GTK_WIDGET_UNSET_FLAGS (viewport, GTK_NO_WINDOW);
    GTK_WIDGET_UNSET_FLAGS (viewport, GTK_CAN_FOCUS);
 
@@ -118,10 +130,18 @@ static void Dw_gtk_viewport_init (GtkDwViewport *viewport)
    viewport->back_pixmap = NULL;
    viewport->child = NULL;
    viewport->last_entered = NULL;
-   viewport->resize_idle_id = -1;
-   viewport->back_pixmap = NULL;
+   viewport->draw_resize_idle_id = 0;
    viewport->anchor = NULL;
-   viewport->anchor_idle_id = -1;
+   viewport->anchor_idle_id = 0;
+   viewport->findtext_state = a_Findtext_state_new ();
+   viewport->selection = a_Selection_new ();
+   viewport->anchors_table = g_hash_table_new (g_str_hash, g_str_equal);
+   viewport->draw_areas = NULL;
+   viewport->num_draw_areas = 0;
+   viewport->num_draw_areas_max = 4;
+   
+   DBG_OBJ_ASSOC (viewport->findtext_state, viewport);
+   DBG_OBJ_ASSOC (viewport->selection, viewport);
 }
 
 
@@ -158,6 +178,14 @@ static void Dw_gtk_viewport_class_init (GtkDwViewportClass *klass)
  *                     *
  ***********************/
 
+static gboolean Dw_gtk_viewport_destroy_anchor (gpointer key,
+                                                gpointer value,
+                                                gpointer user_data)
+{
+   g_free (value);
+   return TRUE;
+}
+
 /*
  * Standard Gtk+ function
  */
@@ -171,14 +199,22 @@ static void Dw_gtk_viewport_destroy (GtkObject *object)
       gdk_pixmap_unref (viewport->back_pixmap);
    if (viewport->child)
       gtk_object_destroy (GTK_OBJECT (viewport->child));
-   if (viewport->resize_idle_id)
-      gtk_idle_remove (viewport->resize_idle_id);
-   if (viewport->anchor)
-      g_free (viewport->anchor);
-   if (viewport->anchor_idle_id != -1)
+   if (viewport->draw_resize_idle_id != 0)
+      gtk_idle_remove (viewport->draw_resize_idle_id);
+   if (viewport->anchor_idle_id != 0)
       gtk_idle_remove (viewport->anchor_idle_id);
+   g_free (viewport->anchor);
 
-   (* GTK_OBJECT_CLASS(parent_class)->destroy) (object);
+   g_hash_table_foreach_remove (viewport->anchors_table,
+                                Dw_gtk_viewport_destroy_anchor, NULL);
+   g_hash_table_destroy (viewport->anchors_table);
+
+   g_free (viewport->draw_areas);
+
+   a_Findtext_state_destroy (viewport->findtext_state);
+   a_Selection_free (viewport->selection);
+
+   GTK_OBJECT_CLASS(parent_class)->destroy (object);
 }
 
 
@@ -195,7 +231,7 @@ static void Dw_gtk_viewport_realize (GtkWidget *widget)
 {
    GtkDwViewport *viewport;
 
-   (* (GTK_WIDGET_CLASS(parent_class)->realize)) (widget);
+   GTK_WIDGET_CLASS(parent_class)->realize (widget);
 
    gdk_window_set_events (widget->window,
                           gdk_window_get_events (widget->window)
@@ -220,7 +256,7 @@ static void Dw_gtk_viewport_unrealize (GtkWidget *widget)
 {
    GtkDwViewport *viewport;
 
-   (* (GTK_WIDGET_CLASS(parent_class)->unrealize)) (widget);
+   GTK_WIDGET_CLASS(parent_class)->unrealize (widget);
 
    viewport = GTK_DW_VIEWPORT (widget);
    if (viewport->child)
@@ -258,7 +294,7 @@ static void Dw_gtk_viewport_paint (GtkWidget *widget,
 
          if (new_back_pixmap) {
             if (viewport->back_pixmap)
-               gdk_pixmap_ref (viewport->back_pixmap);
+               gdk_pixmap_unref (viewport->back_pixmap);
             viewport->back_pixmap = gdk_pixmap_new (widget->window,
                                                     widget->allocation.width,
                                                     widget->allocation.height,
@@ -280,8 +316,7 @@ static void Dw_gtk_viewport_paint (GtkWidget *widget,
          child_area.x = viewport->child->allocation.x;
          child_area.y = viewport->child->allocation.y;
          child_area.width = viewport->child->allocation.width;
-         child_area.height = (viewport->child->allocation.ascent +
-                              viewport->child->allocation.descent);
+         child_area.height = DW_WIDGET_HEIGHT(viewport->child);
 
          if (p_Dw_rectangle_intersect (&parent_area, &child_area,
                                        &intersection)) {
@@ -314,7 +349,7 @@ static void Dw_gtk_viewport_draw (GtkWidget *widget,
                                   GdkRectangle *area)
 {
    Dw_gtk_viewport_paint (widget, area, NULL);
-   (* (GTK_WIDGET_CLASS(parent_class)->draw)) (widget, area);
+   GTK_WIDGET_CLASS(parent_class)->draw (widget, area);
 }
 
 
@@ -325,7 +360,7 @@ static gint Dw_gtk_viewport_expose (GtkWidget *widget,
                                     GdkEventExpose *event)
 {
    Dw_gtk_viewport_paint (widget, &(event->area), event);
-   return (* (GTK_WIDGET_CLASS(parent_class)->expose_event)) (widget, event);
+   return GTK_WIDGET_CLASS(parent_class)->expose_event (widget, event);
 }
 
 
@@ -365,6 +400,11 @@ static gint Dw_gtk_viewport_mouse_event (GtkWidget *widget,
 static gint Dw_gtk_viewport_button_press (GtkWidget *widget,
                                           GdkEventButton *event)
 {
+   /* We focus always the viewport window (i.e., more precisely, the
+    * GktDwScrolledFrame). */
+   if (widget->parent)
+      gtk_widget_grab_focus(widget->parent);
+
    return Dw_gtk_viewport_mouse_event (widget, event->x, event->y,
                                        (GdkEvent*) event);
 }
@@ -445,6 +485,7 @@ void a_Dw_gtk_viewport_add_dw (GtkDwViewport *viewport,
    g_return_if_fail (viewport->child == NULL);
 
    viewport->child = widget;
+   DBG_OBJ_ASSOC(widget, viewport);
 
    widget->parent = NULL;
    widget->viewport = GTK_WIDGET (viewport);
@@ -455,6 +496,9 @@ void a_Dw_gtk_viewport_add_dw (GtkDwViewport *viewport,
 
    Dw_gtk_viewport_calc_size (viewport);
    Dw_gtk_viewport_remove_anchor (viewport);
+
+   a_Findtext_state_set_widget (viewport->findtext_state, widget);
+   a_Selection_reset (viewport->selection);
 }
 
 /**************************************************
@@ -470,6 +514,7 @@ void a_Dw_gtk_viewport_add_dw (GtkDwViewport *viewport,
  */
 void Dw_gtk_viewport_remove_dw (GtkDwViewport *viewport)
 {
+   a_Findtext_state_set_widget (viewport->findtext_state, NULL);
    viewport->child = NULL;
    Dw_gtk_viewport_remove_anchor (viewport);
    Dw_gtk_viewport_calc_size (viewport);
@@ -668,21 +713,27 @@ static void Dw_gtk_viewport_widget_at_point_callback (DwWidget *widget,
    WidgetAtPointData *callback_data;
 
    callback_data = (WidgetAtPointData*) data;
+   DEBUG_MSG (1, "  Checking %p ...\n", widget);
 
-   if (callback_data->x >= widget->allocation.x &&
-       callback_data->y >= widget->allocation.y &&
-       callback_data->x < widget->allocation.x + widget->allocation.width &&
-       callback_data->y < widget->allocation.y + (widget->allocation.ascent +
-                                                  widget->allocation.descent))
-      {
-         if (DW_IS_CONTAINER (widget))
-            a_Dw_container_forall (DW_CONTAINER (widget),
-                                   Dw_gtk_viewport_widget_at_point_callback,
-                                   data);
+   if (/* As a special exception, for the top-level widget, not the
+        * allocation is regarded, but the whole viewport. This makes
+        * selections more useful, since so the user can start the
+        * selection outside of the allocation. */
+       widget->parent == NULL ||
+       /* Otherwise, check whether pointer is in the allocation. */
+       (callback_data->x >= widget->allocation.x &&
+        callback_data->y >= widget->allocation.y &&
+        callback_data->x < widget->allocation.x + widget->allocation.width &&
+        callback_data->y < widget->allocation.y + DW_WIDGET_HEIGHT(widget))) {
+      DEBUG_MSG (1, "    yes\n");
+      if (DW_IS_CONTAINER (widget))
+         a_Dw_container_forall (DW_CONTAINER (widget),
+                                Dw_gtk_viewport_widget_at_point_callback,
+                                data);
 
-         if (callback_data->widget == NULL)
-            callback_data->widget = widget;
-      }
+      if (callback_data->widget == NULL)
+         callback_data->widget = widget;
+   }
 }
 
 /*
@@ -713,47 +764,178 @@ DwWidget* Dw_gtk_viewport_widget_at_point (GtkDwViewport *viewport,
  *************/
 
 /*
- * See Dw_gtk_viewport_scroll_to.
+ * Add or change an anchor.
+ * The widget is responsible for storing a copy of name.
+ * 
+ * todo: Currently, no horizontal scrolling is done. This is generally
+ * possible, DW_HPOS_INTO_VIEW should be used for this, but it is
+ * rather complicated to determine the width of an anchor. This would
+ * be the lenght of the region between <a> and </a>, the page widget
+ * would have to have two kinds of content types (opening and closing
+ * anchor), and some changes in the HTML parser are necessary.
  */
-static gint Dw_gtk_viewport_update_anchor_idle (gpointer data)
+void p_Dw_gtk_viewport_put_anchor (DwWidget *widget,
+                                   gchar *name,
+                                   gint32 y)
 {
    GtkDwViewport *viewport;
-   GtkAdjustment *adj;
+   GtkDwViewportAnchor *anchor;
+   gpointer tmp_anchor;
 
-   viewport = GTK_DW_VIEWPORT (data);
-   adj = GTK_LAYOUT(viewport)->vadjustment;
+   viewport = GTK_DW_VIEWPORT (widget->viewport);
+   _MSG("  name(%p)=[%s]\n", name, name);
+   if (g_hash_table_lookup_extended (viewport->anchors_table, name, NULL,
+                                     &tmp_anchor)) {
+      anchor = tmp_anchor;
+   } else {
+      anchor = g_new (GtkDwViewportAnchor, 1);
+      g_hash_table_insert (viewport->anchors_table, name, anchor);
+   }
 
-   if (viewport->anchor_y > adj->upper - adj->page_size)
-      gtk_adjustment_set_value (adj, adj->upper - adj->page_size);
-   else
-      gtk_adjustment_set_value (adj, viewport->anchor_y);
+   anchor->widget = widget;
+   anchor->y = y;
 
-   viewport->anchor_idle_id = -1;
-   return FALSE;
+   Dw_gtk_viewport_update_anchor (viewport);
 }
 
 
 /*
- * Called by Dw_gtk_viewport_update_anchor.
+ * Used by Dw_gtk_viewport_update_anchor_idle.
  */
-static void Dw_gtk_viewport_update_anchor_rec (DwWidget *widget)
+static gboolean Dw_gtk_viewport_calc_into (gint32 requested_value,
+                                           gint32 requested_size,
+                                           gint32 current_value,
+                                           gint32 size,
+                                           gint32 *return_value)
 {
-   gpointer p;
-   GtkDwViewport *viewport;
-
-   viewport = GTK_DW_VIEWPORT (widget->viewport);
-
-   if (widget->anchors_table &&
-       g_hash_table_lookup_extended (widget->anchors_table, viewport->anchor,
-                                     NULL, &p))
-      Dw_gtk_viewport_scroll_to (viewport,
-                                 GPOINTER_TO_INT(p) + widget->allocation.y);
-   else {
-      if (DW_IS_CONTAINER (widget))
-         a_Dw_container_forall (DW_CONTAINER (widget),
-                                (DwCallback)Dw_gtk_viewport_update_anchor_rec,
-                                NULL);
+   if (requested_size > size) {
+      /* The viewport size is smaller than the size of the region which will
+       * be shown. If the region is already visible, do not change the
+       * position. Otherwise, show the left/upper border, this is most likely
+       * what is needed. */
+      if (current_value >= requested_value &&
+          current_value + size < requested_value + requested_size)
+         return FALSE;
+      else
+         requested_size = size;
    }
+
+   if (requested_value < current_value) {
+      *return_value = requested_value;
+      return TRUE;
+   } else if (requested_value + requested_size > current_value + size) {
+      *return_value = requested_value - size + requested_size;
+      return TRUE;
+   } else
+      return FALSE;
+}
+
+/*
+ * See Dw_gtk_viewport_scroll_to.
+ */
+static gint Dw_gtk_viewport_update_anchor_idle (gpointer data)
+{
+   gint32 vp_width, vp_height, x = 0, y = 0;
+   GtkScrolledWindow *scrolled;
+   GtkDwViewport *viewport;
+   GtkWidget *vwidget;
+   GtkAdjustment *vadj, *hadj;
+   gboolean change_x, change_y;
+
+   DBG_MSG (data, "scrolling", 0, "Dw_gtk_viewport_update_anchor_idle");
+   DBG_MSG_START (data);
+
+   vwidget = GTK_WIDGET (data);
+   viewport = GTK_DW_VIEWPORT (vwidget);
+   scrolled = GTK_SCROLLED_WINDOW (vwidget->parent->parent);
+   hadj = GTK_LAYOUT(viewport)->hadjustment;
+   vadj = GTK_LAYOUT(viewport)->vadjustment;
+
+   vp_width =
+      vwidget->allocation.width - GTK_CONTAINER(viewport)->border_width;
+   vp_height =
+      vwidget->allocation.height - GTK_CONTAINER(viewport)->border_width;
+   DBG_MSGF (viewport, "scrolling", 0, "vp_width = %d", vp_width);
+   DBG_MSGF (viewport, "scrolling", 0, "vp_height = %d", vp_height);
+
+   change_x = TRUE;
+   switch (viewport->anchor_pos.hpos) {
+   case DW_HPOS_LEFT:
+      DBG_MSG (viewport, "scrolling", 0, "DW_HPOS_LEFT");
+      x = viewport->anchor_pos.x;
+      break;
+   case DW_HPOS_CENTER:
+      DBG_MSG (viewport, "scrolling", 0, "DW_HPOS_CENTER");
+      x = viewport->anchor_pos.x - (vp_width - viewport->anchor_pos.width) / 2;
+      break;
+   case DW_HPOS_RIGHT:
+      DBG_MSG (viewport, "scrolling", 0, "DW_HPOS_RIGHT");
+      x = viewport->anchor_pos.x - (vp_width - viewport->anchor_pos.width);
+      break;
+   case DW_HPOS_INTO_VIEW:
+      DBG_MSG (viewport, "scrolling", 0, "DW_HPOS_INTO_VIEW");
+      change_x = Dw_gtk_viewport_calc_into (viewport->anchor_pos.x,
+                                            viewport->anchor_pos.width,
+                                            hadj->value, vp_width, &x);
+      break;
+   case DW_HPOS_NO_CHANGE:
+      DBG_MSG (viewport, "scrolling", 0, "DW_HPOS_NO_CHANGE");
+      change_x = FALSE;
+      break;
+   }
+
+   change_y = TRUE;
+   switch (viewport->anchor_pos.vpos) {
+   case DW_VPOS_TOP:
+      DBG_MSG (viewport, "scrolling", 0, "DW_VPOS_TOP");
+      y = viewport->anchor_pos.y;
+      break;
+   case DW_VPOS_CENTER:
+      DBG_MSG (viewport, "scrolling", 0, "DW_VPOS_CENTER");
+      y = viewport->anchor_pos.y -
+         (vp_height - viewport->anchor_pos.height) / 2;
+      break;
+   case DW_VPOS_BOTTOM:
+      DBG_MSG (viewport, "scrolling", 0, "DW_VPOS_BOTTOM");
+      y = viewport->anchor_pos.y - (vp_height - viewport->anchor_pos.height);
+      break;
+   case DW_VPOS_INTO_VIEW:
+      DBG_MSG (viewport, "scrolling", 0, "DW_VPOS_INTO_VIEW");
+      change_y = Dw_gtk_viewport_calc_into (viewport->anchor_pos.y,
+                                            viewport->anchor_pos.height,
+                                            vadj->value, vp_height, &y);
+   case DW_VPOS_NO_CHANGE:
+      DBG_MSG (viewport, "scrolling", 0, "DW_VPOS_NO_CHANGE");
+      change_y = FALSE;
+      break;
+   }
+
+   DBG_MSGF (viewport, "scrolling", 0, "scrolling to (%d, %d)\n", x, y);
+   DBG_MSGF (viewport, "scrolling", 0,
+             "hadj->upper = %g, hadj->page_size = %g",
+             hadj->upper, hadj->page_size);
+   DBG_MSGF (viewport, "scrolling", 0,
+             "vadj->upper = %g, vadj->page_size = %g",
+             vadj->upper, vadj->page_size);
+
+   if (change_x) {
+      if (x > hadj->upper - hadj->page_size)
+         gtk_adjustment_set_value (hadj, hadj->upper - hadj->page_size);
+      else
+         gtk_adjustment_set_value (hadj, x);
+   }
+
+   if (change_y) {
+      if (y > vadj->upper - vadj->page_size)
+         gtk_adjustment_set_value (vadj, vadj->upper - vadj->page_size);
+      else
+         gtk_adjustment_set_value (vadj, y);
+   }
+
+   viewport->anchor_idle_id = 0;
+
+   DBG_MSG_END (viewport);
+   return FALSE;
 }
 
 
@@ -763,8 +945,17 @@ static void Dw_gtk_viewport_update_anchor_rec (DwWidget *widget)
  */
 void Dw_gtk_viewport_update_anchor (GtkDwViewport *viewport)
 {
-   if (viewport->anchor && viewport->child)
-      Dw_gtk_viewport_update_anchor_rec (viewport->child);
+   GtkDwViewportAnchor *anchor;
+   gpointer tmp_anchor;
+
+   if (viewport->anchor &&
+       g_hash_table_lookup_extended (viewport->anchors_table, viewport->anchor,
+                                     NULL, &tmp_anchor)) {
+      anchor = tmp_anchor;
+      Dw_gtk_viewport_scroll_to (viewport, DW_HPOS_NO_CHANGE, DW_VPOS_TOP,
+                                 0, anchor->y + anchor->widget->allocation.y,
+                                 0, 0);
+   }
 }
 
 
@@ -777,7 +968,7 @@ void a_Dw_gtk_viewport_set_anchor (GtkDwViewport *viewport,
    Dw_gtk_viewport_remove_anchor (viewport);
 
    if (anchor) {
-      viewport->anchor = g_strdup(anchor);
+      viewport->anchor = g_strdup (anchor);
       Dw_gtk_viewport_update_anchor (viewport);
    } else {
       viewport->anchor = NULL;
@@ -790,21 +981,43 @@ void a_Dw_gtk_viewport_set_anchor (GtkDwViewport *viewport,
  * Sets the position to scroll to. The current anchor will be removed.
  */
 void a_Dw_gtk_viewport_set_scrolling_position (GtkDwViewport *viewport,
-                                               gint pos)
+                                               gint32 x,
+                                               gint32 y)
 {
    Dw_gtk_viewport_remove_anchor (viewport);
-   Dw_gtk_viewport_scroll_to (viewport, pos);
+   Dw_gtk_viewport_scroll_to (viewport, DW_HPOS_LEFT, DW_VPOS_TOP, x, y, 0, 0);
 }
 
 /*
- * Scrolls the viewport to position y.
- * This is done in an idle function.
+ * Scrolls the viewport, so that the region [x, y, width, height] (world
+ * coordinates) is seen, according to hpos and vpos.
+ *
+ * The actual scrolling is done in an idle function.
  */
 void Dw_gtk_viewport_scroll_to (GtkDwViewport *viewport,
-                                gint32 y)
+                                DwHPosition hpos,
+                                DwVPosition vpos,
+                                gint32 x,
+                                gint32 y,
+                                gint32 width,
+                                gint32 height)
 {
-   viewport->anchor_y = y;
-   if (viewport->anchor_idle_id == -1)
+   viewport->anchor_pos.hpos = hpos;
+   viewport->anchor_pos.vpos = vpos;
+   viewport->anchor_pos.x = x;
+   viewport->anchor_pos.y = y;
+   viewport->anchor_pos.width = width;
+   viewport->anchor_pos.height = height;
+
+   DBG_OBJ_SET_NUM (viewport, "anchor_pos.hpos", viewport->anchor_pos.hpos);
+   DBG_OBJ_SET_NUM (viewport, "anchor_pos.vpos", viewport->anchor_pos.vpos);
+   DBG_OBJ_SET_NUM (viewport, "anchor_pos.x", viewport->anchor_pos.x);
+   DBG_OBJ_SET_NUM (viewport, "anchor_pos.y", viewport->anchor_pos.y);
+   DBG_OBJ_SET_NUM (viewport, "anchor_pos.width", viewport->anchor_pos.width);
+   DBG_OBJ_SET_NUM (viewport, "anchor_pos.height",
+                    viewport->anchor_pos.height);
+
+   if (viewport->anchor_idle_id == 0)
       viewport->anchor_idle_id = gtk_idle_add
          (Dw_gtk_viewport_update_anchor_idle, (gpointer)viewport);
 }
@@ -820,8 +1033,147 @@ void Dw_gtk_viewport_remove_anchor (GtkDwViewport *viewport)
       viewport->anchor = NULL;
    }
 
-   if (viewport->anchor_idle_id != -1) {
+   if (viewport->anchor_idle_id != 0) {
       gtk_idle_remove (viewport->anchor_idle_id);
-      viewport->anchor_idle_id = -1;
+      viewport->anchor_idle_id = 0;
    }
 }
+
+/*
+ * Drawing and resizing is done in this idle function.
+ */
+static gint Dw_gtk_viewport_draw_resize_idle (gpointer data)
+{
+   GtkDwViewport *viewport;
+   GtkLayout *layout;
+   GtkWidget *widget;
+   DwRectangle viewport_area, world_area;
+   GdkRectangle gtk_area;
+   int i;
+
+   viewport = GTK_DW_VIEWPORT (data);
+
+   switch (viewport->draw_resize_action) {
+   case DW_GTK_VIEWPORT_DRAW:
+      for (i = 0; i < viewport->num_draw_areas; i++) {
+         widget = GTK_WIDGET (viewport);
+         layout = GTK_LAYOUT (viewport);
+
+         viewport_area.x = gtk_layout_get_hadjustment(layout)->value;
+         viewport_area.y = gtk_layout_get_vadjustment(layout)->value;;
+         viewport_area.width = widget->allocation.width;
+         viewport_area.height = widget->allocation.height;
+      
+         if (p_Dw_rectangle_intersect (&viewport->draw_areas[i],
+                                       &viewport_area, &world_area)) {
+            gtk_area.x = world_area.x - viewport_area.x;
+            gtk_area.y = world_area.y - viewport_area.y;
+            gtk_area.width = world_area.width;
+            gtk_area.height = world_area.height;
+            gtk_widget_draw (widget, &gtk_area);
+         }
+      }
+      
+      /* No more areas to be drawn. */
+      viewport->num_draw_areas = 0;
+      break;
+      
+   case DW_GTK_VIEWPORT_RESIZE:
+      Dw_gtk_viewport_calc_size (viewport);
+      break;
+   }
+   
+   viewport->draw_resize_idle_id = 0;
+   return FALSE;
+}
+
+/*
+ * Queue an area for drawing. This function is called by
+ * p_Dw_widget_queue_draw_area. x and y are passed in world coordinates.
+ */
+void Dw_gtk_viewport_queue_draw (GtkDwViewport *viewport,
+                                 gint32 x,
+                                 gint32 y,
+                                 gint32 width,
+                                 gint32 height)
+{
+   DwRectangle area;
+   int i;
+
+   if (viewport->draw_resize_idle_id == 0) {
+      viewport->draw_resize_action = DW_GTK_VIEWPORT_DRAW;
+      viewport->draw_resize_idle_id =
+         gtk_idle_add (Dw_gtk_viewport_draw_resize_idle, (gpointer)viewport);
+   } else if (viewport->draw_resize_action == DW_GTK_VIEWPORT_RESIZE)
+      /* Drawing is always overridden by resizing. */
+      return;
+
+   area.x = x;
+   area.y = y;
+   area.width = width;
+   area.height = height;
+
+   /* First, try to keep the list as clean as possible. Check whether other
+    * rectangles interfer with this one in some way. */
+   /* An idea for optimization: The list could be sorted, and so the part of
+    * the list we have to consider here, may be reduced, the start may be
+    * found via linear search. However, this probably makes balanced binary
+    * trees necessary, since moving elements within the array may be quite
+    * time-consuming.
+    */
+   _MSG("  num_draw_areas = %d\n", viewport->num_draw_areas);
+   for (i = 0; i < viewport->num_draw_areas; i++) {
+      if (p_Dw_rectangle_is_subset (&area, &viewport->draw_areas[i]))
+         /* First case: area is a subset of an already queued rectangle
+          * -> nothing to do. */
+         return;
+      else if (p_Dw_rectangle_is_subset (&viewport->draw_areas[i], &area)) {
+         /* Second case: area is a subset of an already queued rectangle
+          * -> replace the other one with area. */
+         viewport->draw_areas[i] = area;
+         return;
+      }
+      /* Maybe some more tests: if both areas may exactly be combined to a
+       * rectangle? Very unlikely case ... */
+   }
+
+   /* No interference: add  the new area to the list. */
+   viewport->num_draw_areas++;
+   a_List_add (viewport->draw_areas, viewport->num_draw_areas,
+               viewport->num_draw_areas_max);
+   viewport->draw_areas[viewport->num_draw_areas - 1] = area;
+}
+
+/*
+ * Start the resizing idle. This function is called by
+ * p_Dw_widget_queue_resize, after the appropriate attributes have been set in
+ * the widgets, where necessary.
+ */
+void Dw_gtk_viewport_queue_resize (GtkDwViewport *viewport)
+{
+   /* Resizing always overrides drawing. */
+   viewport->draw_resize_action = DW_GTK_VIEWPORT_RESIZE;
+   viewport->num_draw_areas = 0;
+   
+   if (viewport->draw_resize_idle_id == 0)
+      viewport->draw_resize_idle_id =
+         gtk_idle_add (Dw_gtk_viewport_draw_resize_idle, (gpointer)viewport);
+}
+
+
+/*
+ * Return the DwWidget which is at position (vx, vy) in viewport coordinates.
+ */
+DwWidget* a_Dw_gtk_viewport_widget_at_viewport_point (GtkDwViewport *viewport,
+                                                      gint32 vx,
+                                                      gint32 vy)
+{
+   gint32 world_x, world_y;
+
+   if (viewport->child) {
+      world_x = vx + gtk_layout_get_hadjustment(GTK_LAYOUT(viewport))->value;
+      world_y = vy + gtk_layout_get_vadjustment(GTK_LAYOUT(viewport))->value;
+      return Dw_gtk_viewport_widget_at_point (viewport, world_x, world_y);
+   } else
+      return NULL;   
+}
