@@ -2,7 +2,7 @@
  * File: dw_image.c
  *
  * Copyright (C) 2001 Sebastian Geerken  <S.Geerken@ping.de>,
- *                    Jorge Arellano Cid <jcid@inf.utfsm.cl>
+ *                    Jorge Arellano Cid <jcid@dillo.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -10,12 +10,14 @@
  * (at your option) any later version.
  */
 
+#include "msg.h"
 #include "dw_image.h"
 #include "dw_gtk_viewport.h"
 #include "prefs.h"
 #include "dw_marshal.h"
 #include "list.h"
 #include "dicache.h"
+#include "debug.h"
 #include <gdk/gdk.h>
 #include <gtk/gtk.h>
 #include <stdio.h>
@@ -33,24 +35,31 @@ static void Dw_image_size_allocate      (DwWidget *widget,
 static void Dw_image_draw               (DwWidget *widget,
                                          DwRectangle *area,
                                          GdkEventExpose *event);
-static gint Dw_image_button_press       (DwWidget *widget,
+static gboolean Dw_image_button_press   (DwWidget *widget,
                                          gint32 x,
                                          gint32 y,
                                          GdkEventButton *event);
-static gint Dw_image_button_release     (DwWidget *widget,
+static gboolean Dw_image_button_release (DwWidget *widget,
                                          gint32 x,
                                          gint32 y,
                                          GdkEventButton *event);
-static gint Dw_image_motion_notify      (DwWidget *widget,
+static gboolean Dw_image_motion_notify  (DwWidget *widget,
                                          gint32 x,
                                          gint32 y,
                                          GdkEventMotion *event);
-static gint Dw_image_enter_notify       (DwWidget *widget,
+static gboolean Dw_image_enter_notify   (DwWidget *widget,
                                          DwWidget *last_widget,
                                          GdkEventMotion *event);
-static gint Dw_image_leave_notify       (DwWidget *widget,
+static gboolean Dw_image_leave_notify   (DwWidget *widget,
                                          DwWidget *next_widget,
                                          GdkEventMotion *event);
+static DwIterator* Dw_image_iterator    (DwWidget *widget,
+                                         gint32 mask,
+                                         gboolean at_end);
+static void Dw_image_iterator_highlight (DwIterator *it,
+                                         gint start,
+                                         gint end,
+                                         DwHighlightLayer layer);
 
 static void Dw_image_find_link          (DwImage *image,
                                          gint x, gint y,
@@ -114,13 +123,15 @@ GtkType a_Dw_image_get_type (void)
 /*
  * Standard Gtk+ function.
  */
-DwWidget* a_Dw_image_new (DwImageType type, const gchar *alt)
+DwWidget* a_Dw_image_new (DwImageType type, const gchar *alt_text)
 {
    GtkObject *object;
+   DwImage *image;
 
    object = gtk_object_new (DW_TYPE_IMAGE, NULL);
-   if (alt && alt[0] && prefs.show_alt)
-      DW_IMAGE(object)->alt_tooltip = a_Dw_tooltip_new (alt);
+   DBG_OBJ_CREATE (object, "DwImage");
+   image = DW_IMAGE(object);
+   image->alt_text = g_strdup (alt_text);
    return DW_WIDGET (object);
 }
 
@@ -130,13 +141,19 @@ DwWidget* a_Dw_image_new (DwImageType type, const gchar *alt)
  */
 static void Dw_image_init (DwImage *image)
 {
+   int i;
+
    image->url = NULL;
    image->width = 0;
    image->height = 0;
+   image->alt_text_width = -1; /* not yet calculated */
    image->buffer = NULL;
    image->scaled_buffer = NULL;
-   image->alt_tooltip = NULL;
+   image->alt_text = NULL;
    image->usemap_url = NULL;
+   image->ismap = FALSE;
+   for (i = 0; i < DW_HIGHLIGHT_NUM_LAYERS; i++)
+      image->selected[i] = FALSE;
 }
 
 
@@ -155,37 +172,37 @@ static void Dw_image_class_init (DwImageClass *klass)
 
    image_signals[LINK_ENTERED] =
       gtk_signal_new ("link_entered",
-                      GTK_RUN_FIRST,
+                      GTK_RUN_LAST,
                       object_class->type,
                       GTK_SIGNAL_OFFSET (DwImageClass, link_entered),
                       p_Dw_marshal_link_enter,
-                      GTK_TYPE_NONE,
+                      GTK_TYPE_BOOL,
                       3, GTK_TYPE_INT, GTK_TYPE_INT, GTK_TYPE_INT);
    image_signals[LINK_PRESSED] =
       gtk_signal_new ("link_pressed",
-                      GTK_RUN_FIRST,
+                      GTK_RUN_LAST,
                       object_class->type,
                       GTK_SIGNAL_OFFSET (DwImageClass, link_pressed),
                       p_Dw_marshal_link_button,
-                      GTK_TYPE_NONE,
+                      GTK_TYPE_BOOL,
                       4, GTK_TYPE_INT, GTK_TYPE_INT, GTK_TYPE_INT,
                       GTK_TYPE_GDK_EVENT);
    image_signals[LINK_RELEASED] =
       gtk_signal_new ("link_released",
-                      GTK_RUN_FIRST,
+                      GTK_RUN_LAST,
                       object_class->type,
                       GTK_SIGNAL_OFFSET (DwImageClass, link_released),
                       p_Dw_marshal_link_button,
-                      GTK_TYPE_NONE,
+                      GTK_TYPE_BOOL,
                       4, GTK_TYPE_INT, GTK_TYPE_INT, GTK_TYPE_INT,
                       GTK_TYPE_GDK_EVENT);
    image_signals[LINK_CLICKED] =
       gtk_signal_new ("link_clicked",
-                      GTK_RUN_FIRST,
+                      GTK_RUN_LAST,
                       object_class->type,
                       GTK_SIGNAL_OFFSET (DwImageClass, link_clicked),
                       p_Dw_marshal_link_button,
-                      GTK_TYPE_NONE,
+                      GTK_TYPE_BOOL,
                       4, GTK_TYPE_INT, GTK_TYPE_INT, GTK_TYPE_INT,
                       GTK_TYPE_GDK_EVENT);
    gtk_object_class_add_signals (object_class, image_signals, LAST_SIGNAL);
@@ -199,6 +216,7 @@ static void Dw_image_class_init (DwImageClass *klass)
    widget_class->motion_notify_event = Dw_image_motion_notify;
    widget_class->enter_notify_event = Dw_image_enter_notify;
    widget_class->leave_notify_event = Dw_image_leave_notify;
+   widget_class->iterator = Dw_image_iterator;
 }
 
 
@@ -209,16 +227,14 @@ static void Dw_image_destroy (GtkObject *object)
 {
    DwImage *image = DW_IMAGE (object);
 
-   if (image->alt_tooltip)
-      a_Dw_tooltip_destroy (image->alt_tooltip);
-   if (image->scaled_buffer)
-      g_free (image->scaled_buffer);
    if (image->usemap_url)
-      a_Url_free(image->usemap_url);
+      a_Url_free (image->usemap_url);
    if (image->url)
-      a_Dicache_unref(image->url, image->version);
+      a_Dicache_unref (image->url, image->version);
+   g_free (image->alt_text);
+   g_free (image->scaled_buffer);
 
-   (* GTK_OBJECT_CLASS(parent_class)->destroy) (object);
+   GTK_OBJECT_CLASS(parent_class)->destroy (object);
 }
 
 
@@ -231,11 +247,25 @@ static void Dw_image_size_request (DwWidget *widget,
    DwImage *image;
 
    image = DW_IMAGE (widget);
-   requisition->width =
-      image->width + p_Dw_style_box_diff_width (widget->style);
-   requisition->ascent =
-      image->height + p_Dw_style_box_diff_height (widget->style);
-   requisition->descent = 0;
+   
+   if (image->buffer != NULL ||
+       (image->alt_text == NULL || image->alt_text[0] == '\0')) {
+      requisition->width = image->width;
+      requisition->ascent = image->height;
+      requisition->descent = 0;
+   } else {
+      if (image->alt_text_width == -1)
+         image->alt_text_width = 
+            gdk_string_width (widget->style->font->font, image->alt_text);
+
+      requisition->width = image->alt_text_width;
+      requisition->ascent = widget->style->font->font->ascent;
+      requisition->descent = widget->style->font->font->descent;
+   }
+
+   requisition->width += p_Dw_style_box_diff_width (widget->style);
+   requisition->ascent += p_Dw_style_box_offset_y (widget->style);
+   requisition->descent += p_Dw_style_box_rest_height (widget->style);
 }
 
 
@@ -249,8 +279,7 @@ static void Dw_image_size_allocate (DwWidget *widget,
 
    /* if image is moved only */
    if (allocation->width == widget->allocation.width &&
-       allocation->ascent + allocation->descent
-       == widget->allocation.ascent + widget->allocation.descent)
+       allocation->ascent + allocation->descent == DW_WIDGET_HEIGHT(widget))
       return;
 
    /* this is also done in a_Dw_widget_size_allocate, but
@@ -275,19 +304,16 @@ static void Dw_image_draw (DwWidget *widget,
    GdkGC *gc;
    DwImage *image = DW_IMAGE (widget);
    guchar *buffer, *bstart;
-
-   //g_print(">Area x=%d y=%d w=%d h=%d\n", area->x, area->y,
-   //        area->width, area->height);
-
-   p_Dw_widget_draw_widget_box (widget, area);
+   int i;
+   gboolean selected = FALSE;
 
    if (image->buffer) {
+      p_Dw_widget_draw_widget_box (widget, area, FALSE);
+
       content.x = p_Dw_style_box_offset_x (widget->style);
       content.y = p_Dw_style_box_offset_y (widget->style);
-      content.width =
-         widget->allocation.width - p_Dw_style_box_diff_width (widget->style);
-      content.height = widget->allocation.ascent + widget->allocation.descent -
-         p_Dw_style_box_diff_height (widget->style);
+      content.width = DW_WIDGET_CONTENT_WIDTH(widget);
+      content.height = DW_WIDGET_CONTENT_HEIGHT(widget);
 
       if (p_Dw_rectangle_intersect (area, &content, &intersection)) {
          vx = p_Dw_widget_x_world_to_viewport (widget, widget->allocation.x);
@@ -304,27 +330,49 @@ static void Dw_image_draw (DwWidget *widget,
                   content.width * (intersection.y -
                                    p_Dw_style_box_offset_y (widget->style)) );
 
-         gdk_draw_rgb_image(
+         gdk_draw_rgb_image (
             DW_WIDGET_WINDOW (widget), gc,
             vx + intersection.x, vy + intersection.y,
             intersection.width, intersection.height, GDK_RGB_DITHER_MAX,
             bstart, content.width * 3);
       }
+   } else {
+      if (image->alt_text && image->alt_text[0]) {
+         if (image->alt_text_width == -1)
+            image->alt_text_width = 
+               gdk_string_width (widget->style->font->font, image->alt_text);
+
+         if (widget->allocation.width < image->alt_text_width ||
+             widget->allocation.ascent + widget->allocation.descent
+             < widget->style->font->font->ascent
+             + widget->style->font->font->descent)
+            p_Dw_widget_will_clip (widget);
+
+         p_Dw_widget_draw_widget_box (widget, area, FALSE);
+         gdk_draw_string
+            (DW_WIDGET_WINDOW (widget), widget->style->font->font,
+             widget->style->color->gc,
+             p_Dw_widget_x_world_to_viewport (widget, widget->allocation.x),
+             p_Dw_widget_y_world_to_viewport (widget, widget->allocation.y)
+             + widget->style->font->font->ascent,
+             image->alt_text);
+      }
    }
+
+   for (i = 0; i < DW_HIGHLIGHT_NUM_LAYERS && !selected; i++)
+      selected = image->selected[i];
+   if (selected)
+      p_Dw_widget_draw_selected (widget, area);
 }
 
 
 /*
  * Standard Dw function.
  */
-static gint Dw_image_enter_notify (DwWidget *widget,
-                                   DwWidget *last_widget,
-                                   GdkEventMotion *event)
+static gboolean Dw_image_enter_notify (DwWidget *widget,
+                                       DwWidget *last_widget,
+                                       GdkEventMotion *event)
 {
-   DwImage *image = DW_IMAGE (widget);
-
-   if (image->alt_tooltip)
-      a_Dw_tooltip_on_enter (image->alt_tooltip);
    return FALSE;
 }
 
@@ -332,19 +380,18 @@ static gint Dw_image_enter_notify (DwWidget *widget,
 /*
  * Standard Dw function.
  */
-static gint Dw_image_leave_notify (DwWidget *widget,
-                                   DwWidget *next_widget,
-                                   GdkEventMotion *event)
+static gboolean Dw_image_leave_notify (DwWidget *widget,
+                                       DwWidget *next_widget,
+                                       GdkEventMotion *event)
 {
    DwImage *image = DW_IMAGE (widget);
-
-   if (image->alt_tooltip)
-      a_Dw_tooltip_on_leave (image->alt_tooltip);
+   gboolean return_val = FALSE;
 
    if (image->hover_link != -1) {
       image->hover_link = -1;
       gtk_signal_emit (GTK_OBJECT (widget), image_signals[LINK_ENTERED],
-                       -1, -1, -1);
+                       -1, -1, -1, &return_val);
+      return return_val;
    }
 
    return FALSE;
@@ -354,35 +401,37 @@ static gint Dw_image_leave_notify (DwWidget *widget,
 /*
  * Standard Dw function.
  */
-static gint Dw_image_button_press (DwWidget *widget,
-                                   gint32 x,
-                                   gint32 y,
-                                   GdkEventButton *event)
+static gboolean Dw_image_button_press (DwWidget *widget,
+                                       gint32 x,
+                                       gint32 y,
+                                       GdkEventButton *event)
 {
    DwImage *image = DW_IMAGE (widget);
    gint link_x, link_y;
+   gboolean return_val = FALSE;
 
    Dw_image_find_link (image, x, y, &image->pressed_link, &link_x, &link_y);
 
-   if (image->pressed_link >= 0) {
+   if (image->pressed_link >= 0)
       gtk_signal_emit (GTK_OBJECT (widget), image_signals[LINK_PRESSED],
-                       image->pressed_link, link_x, link_y, event);
-      return TRUE;
-   } else
-      return FALSE;
+                       image->pressed_link, link_x, link_y, event,
+                       &return_val);
+
+   return return_val;
 }
 
 
 /*
  * Standard Dw function.
  */
-static gint Dw_image_button_release (DwWidget *widget,
-                                     gint32 x,
-                                     gint32 y,
-                                     GdkEventButton *event)
+static gboolean Dw_image_button_release (DwWidget *widget,
+                                         gint32 x,
+                                         gint32 y,
+                                         GdkEventButton *event)
 {
    DwImage *image = DW_IMAGE (widget);
    gint link_pressed, link_released, link_x, link_y;
+   gboolean return_val1 = FALSE, return_val2 = FALSE;
 
    link_pressed = image->pressed_link;
    Dw_image_find_link (image, x, y, &link_released, &link_x, &link_y);
@@ -390,28 +439,30 @@ static gint Dw_image_button_release (DwWidget *widget,
 
    if (link_released >= 0) {
       gtk_signal_emit (GTK_OBJECT (widget), image_signals[LINK_RELEASED],
-                       link_released, link_x, link_y, event);
-      if (link_pressed == link_released)
+                       link_released, link_x, link_y, event, &return_val1);
+      if (link_pressed == link_released) {
          gtk_signal_emit (GTK_OBJECT (widget), image_signals[LINK_CLICKED],
-                          link_released, link_x, link_y, event);
-      return TRUE;
-   } else
-      return FALSE;
+                          link_released, link_x, link_y, event, &return_val2);
+         return return_val1 || return_val2;
+      }
+      
+      return return_val1;
+   }
+
+   return FALSE;
 }
 
 
 /*
  * Standard Dw function.
  */
-static gint Dw_image_motion_notify (DwWidget *widget,
-                                    gint32 x, gint32 y,
-                                    GdkEventMotion *event)
+static gboolean Dw_image_motion_notify (DwWidget *widget,
+                                        gint32 x, gint32 y,
+                                        GdkEventMotion *event)
 {
    DwImage *image = DW_IMAGE (widget);
    gint link, link_old, link_x, link_y;
-
-   if (image->alt_tooltip)
-      a_Dw_tooltip_on_motion (image->alt_tooltip);
+   gboolean return_val = FALSE;
 
    link_old = image->hover_link;
    Dw_image_find_link (image, x, y, &link, &link_x, &link_y);
@@ -420,8 +471,8 @@ static gint Dw_image_motion_notify (DwWidget *widget,
    /* In case of a ISMAP image, the signal has to be emitted always */
    if (link_x != -1 || link != link_old) {
       gtk_signal_emit (GTK_OBJECT (widget), image_signals[LINK_ENTERED],
-                       link, link_x, link_y);
-      return TRUE;
+                       link, link_x, link_y, &return_val);
+      return return_val;
    } else
       return (link != -1);
 }
@@ -442,15 +493,17 @@ static void Dw_image_find_link (DwImage *image,
    *link_y = -1;
 
    if (image->map_list && image->usemap_url) {
-      *link = Dw_image_map_list_find_link (image->map_list,
-                                           image->usemap_url, x, y);
+      *link = Dw_image_map_list_find_link
+         (image->map_list, image->usemap_url,
+          x - p_Dw_style_box_offset_x (widget->style),
+          y - p_Dw_style_box_offset_y (widget->style));
       if (*link != -1)
          return;
    }
 
    if (image->ismap &&
-       (style = DW_WIDGET(image)->style) && style->link != -1) {
-      *link = style->link;
+       (style = DW_WIDGET(image)->style) && style->x_link != -1) {
+      *link = style->x_link;
       *link_x = x - p_Dw_style_box_offset_x (widget->style);
       *link_y = y - p_Dw_style_box_offset_y (widget->style);
       return;
@@ -463,14 +516,14 @@ static void Dw_image_find_link (DwImage *image,
 /*
  * Set or resize a image.
  */
-void a_Dw_image_size(DwImage *image, gint width, gint height)
+void a_Dw_image_size (DwImage *image, gint width, gint height)
 {
    gint Resize = (image->width != width || image->height != height);
 
    image->width = width;
    image->height = height;
-   if ( Resize )
-      p_Dw_widget_queue_resize(DW_WIDGET (image), 0, TRUE);
+   if (Resize)
+      p_Dw_widget_queue_resize (DW_WIDGET (image), 0, TRUE);
 
    if (image->buffer)
       /* if a_Dw_image_set_buffer has been called before */
@@ -481,13 +534,13 @@ void a_Dw_image_size(DwImage *image, gint width, gint height)
  * Called after the RGB line buffer has been copied into the full
  * image buffer. Uses for drawing and scaling.
  */
-void a_Dw_image_draw_row(DwImage *image,
-                         gint Width, gint Height, gint x, gint y)
+void a_Dw_image_draw_row (DwImage *image,
+                          gint Width, gint Height, gint x, gint y)
 {
    DwWidget *widget = DW_WIDGET (image);
    gint dy1, dy2;
 
-   // g_print("a_Dw_image_draw_row: x=%d y=%d\n", x, y);
+   _MSG("a_Dw_image_draw_row: x=%d y=%d\n", x, y);
    g_return_if_fail (image->buffer != NULL);
 
    if (image->scaled_buffer) {
@@ -509,8 +562,8 @@ void a_Dw_image_draw_row(DwImage *image,
 /*
  * Set the widget buffer to reference the dicache entry buffer
  */
-void a_Dw_image_set_buffer(DwImage *image, guchar *ImageBuffer,
-                           DilloUrl *url, gint version)
+void a_Dw_image_set_buffer (DwImage *image, guchar *ImageBuffer,
+                            DilloUrl *url, gint version)
 {
    image->buffer = ImageBuffer;
    image->url = url;
@@ -538,7 +591,7 @@ void a_Dw_image_set_usemap (DwImage *image,  DwImageMapList *map_list,
    image->map_list = map_list;
 
    if (image->usemap_url)
-      a_Url_free(image->usemap_url);
+      a_Url_free (image->usemap_url);
    image->usemap_url = usemap_url ? a_Url_dup (usemap_url) : NULL;
 }
 
@@ -558,9 +611,8 @@ static void Dw_image_scale (DwImage *image)
    }
 
    widget = DW_WIDGET (image);
-   w = widget->allocation.width - p_Dw_style_box_diff_width (widget->style);
-   h = widget->allocation.ascent + widget->allocation.descent -
-      p_Dw_style_box_diff_height (widget->style);
+   w = DW_WIDGET_CONTENT_WIDTH(widget);
+   h = DW_WIDGET_CONTENT_HEIGHT(widget);
 
    /* Zero or negative sizes? Ignore. */
    if (w <= 0 || h <= 0)
@@ -587,8 +639,7 @@ static void Dw_image_scale_row (DwImage *image, gint y_src)
 
    widget = DW_WIDGET (image);
    w_src = image->width;
-   w_dest =
-      widget->allocation.width - p_Dw_style_box_diff_width (widget->style);
+   w_dest = DW_WIDGET_CONTENT_WIDTH(widget);
    y_dest1 = Dw_image_scaled_y (image, y_src);
    y_dest2 = Dw_image_scaled_y (image, y_src + 1);
 
@@ -649,11 +700,11 @@ void a_Dw_image_map_list_init (DwImageMapList *list)
 {
    list->num_maps = 0;
    list->num_maps_max = 8;
-   list->maps = g_new(DwImageMap, list->num_maps_max);
+   list->maps = g_new (DwImageMap, list->num_maps_max);
 
    list->num_shapes = 0;
    list->num_shapes_max = 8;
-   list->shapes = g_new(DwImageMapShape, list->num_shapes_max);
+   list->shapes = g_new (DwImageMapShape, list->num_shapes_max);
 }
 
 
@@ -684,7 +735,7 @@ void a_Dw_image_map_list_add_map (DwImageMapList *list,
                                   DilloUrl *url)
 {
    list->num_maps++;
-   a_List_add(list->maps, list->num_maps, list->num_maps_max);
+   a_List_add (list->maps, list->num_maps, list->num_maps_max);
    list->maps[list->num_maps - 1].url = a_Url_dup (url);
    list->maps[list->num_maps - 1].start_shape = list->num_shapes;
 }
@@ -718,7 +769,7 @@ void a_Dw_image_map_list_add_shape (DwImageMapList *list,
 
    if (correct_args) {
       list->num_shapes++;
-      a_List_add(list->shapes, list->num_shapes, list->num_shapes_max);
+      a_List_add (list->shapes, list->num_shapes, list->num_shapes_max);
       shape = &list->shapes[list->num_shapes - 1];
       shape->type = type;
       shape->link = link;
@@ -800,3 +851,32 @@ static gint Dw_image_map_list_find_link (DwImageMapList *list,
    /* no map found */
    return -1;
 }
+
+static DwIterator *Dw_image_iterator (DwWidget *widget,
+                                       gint32 mask,
+                                       gboolean at_end)
+{
+   DwIterator *it;
+   DwImage *image = DW_IMAGE (widget);
+
+   if (image->alt_text) {
+      it = p_Dw_widget_text_iterator (widget, mask, at_end, image->alt_text);
+      if (it)
+         it->highlight = Dw_image_iterator_highlight;
+      return it;
+   } else
+      return NULL;
+}
+
+static void Dw_image_iterator_highlight (DwIterator *it,
+                                          gint start,
+                                          gint end,
+                                          DwHighlightLayer layer)
+{
+   if (it->content.type == DW_CONTENT_TEXT) {
+      /* The whole image is highlighted, as soon as something is selected at
+       * all. */
+      DW_IMAGE(it->widget)->selected[layer] = (start != -1 && start != end);
+      p_Dw_widget_queue_draw (it->widget);
+   }
+}
