@@ -1,7 +1,7 @@
 /*
  * File: dw_table.c
  *
- * Copyright (C) 2001 Sebastian Geerken <sgeerken@users.sourceforge.net>
+ * Copyright (C) 2001-2003 Sebastian Geerken <sgeerken@users.sourceforge.net>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,11 +16,12 @@
 #include "prefs.h"
 #include "dw_gtk_viewport.h"
 
-#define DEBUG_CALC_LEVEL   0
-#define DEBUG_EXTR_LEVEL   0
-#define DEBUG_WIDTH_LEVEL 10
+#define DEBUG_CALC_LEVEL      0
+#define DEBUG_EXTR_LEVEL     10
+#define DEBUG_WIDTH_LEVEL     0
+#define DEBUG_ITERATOR_LEVEL  0
 
-//#define DEBUG_LEVEL 10
+/*#define DEBUG_LEVEL 10*/
 #include "debug.h"
 
 /*
@@ -65,6 +66,18 @@ static void Dw_table_set_descent   (DwWidget *widget,
 static void Dw_table_draw          (DwWidget *widget,
                                     DwRectangle *area,
                                     GdkEventExpose *event);
+static gboolean Dw_table_button_press  (DwWidget *widget,
+                                        gint32 x,
+                                        gint32 y,
+                                        GdkEventButton *event);
+static gboolean Dw_table_button_release(DwWidget *widget,
+                                        gint32 x,
+                                        gint32 y,
+                                        GdkEventButton *event);
+static gboolean Dw_table_motion_notify (DwWidget *widget,
+                                        gint32 x,
+                                        gint32 y,
+                                        GdkEventMotion *event);
 
 static void Dw_table_add           (DwContainer *container,
                                     DwWidget *widget);
@@ -73,9 +86,13 @@ static void Dw_table_remove        (DwContainer *container,
 static void Dw_table_forall        (DwContainer *container,
                                     DwCallback callback,
                                     gpointer callback_data);
-static gint Dw_table_findtext      (DwContainer *container,
-                                    gpointer FP, gpointer KP,
-                                    gchar *NewKey);
+
+static DwIterator* Dw_table_iterator          (DwWidget *widget,
+                                               gint mask,
+                                               gboolean at_end);
+static gboolean    Dw_table_iterator_next     (DwIterator *it);
+static gboolean    Dw_table_iterator_prev     (DwIterator *it);
+
 
 static void Dw_table_realloc_children (DwTable *table);
 
@@ -137,6 +154,7 @@ DwWidget* a_Dw_table_new (void)
    GtkObject *object;
 
    object = gtk_object_new (DW_TYPE_TABLE, NULL);
+   DBG_OBJ_CREATE (object, "DwTable");
    return DW_WIDGET (object);
 }
 
@@ -169,6 +187,9 @@ static void Dw_table_init (DwTable *table)
    table->row_style_max = 1;
    table->row_style = g_new(DwStyle*, table->row_style_max);
 
+   table->baseline_max = 1;
+   table->baseline = g_new(gint32, table->baseline_max);
+
    table->sub = NULL;
 }
 
@@ -197,11 +218,14 @@ static void Dw_table_class_init (DwTableClass *klass)
    widget_class->set_ascent = Dw_table_set_ascent;
    widget_class->set_descent = Dw_table_set_descent;
    widget_class->draw = Dw_table_draw;
+   widget_class->button_press_event = Dw_table_button_press;
+   widget_class->button_release_event = Dw_table_button_release;
+   widget_class->motion_notify_event = Dw_table_motion_notify;
+   widget_class->iterator = Dw_table_iterator;
 
    container_class->add = Dw_table_add;
    container_class->remove = Dw_table_remove;
    container_class->forall = Dw_table_forall;
-   container_class->findtext = Dw_table_findtext;
 }
 
 
@@ -229,14 +253,14 @@ static void Dw_table_destroy (GtkObject *object)
       if (table->row_style[i])
          a_Dw_style_unref (table->row_style[i]);
    g_free (table->row_style);
+   g_free (table->baseline);
 
    if (table->sub)
       Dw_table_sub_free (table->sub);
 
-   (* GTK_OBJECT_CLASS(parent_class)->destroy) (object);
+   GTK_OBJECT_CLASS(parent_class)->destroy (object);
 }
 
-
 /*
  * Standard Dw function.
  * Furthermore, table->width and table->cum_height are filled.
@@ -256,24 +280,23 @@ static void Dw_table_size_request (DwWidget *widget,
    if (table->num_cols > 0) {
       a_Dw_widget_get_extremes (DW_WIDGET (table), &extremes);
 
-      if (DW_STYLE_IS_LENGTH (widget->style->width))
-         avail_width = DW_STYLE_GET_LENGTH (widget->style->width,
-                                            widget->style->font);
-      else if (DW_STYLE_IS_PERCENTAGE (widget->style->width))
+      if (DW_STYLE_IS_ABS_LENGTH (widget->style->width))
+         avail_width = DW_STYLE_ABS_LENGTH_VAL (widget->style->width);
+      else if (DW_STYLE_IS_PER_LENGTH (widget->style->width))
          /*
           * If the width is > 100%, we use 100%, this prevents ugly
           * results. (May be changed in future, when a more powerful
           * rendering is implemented, to handle fixed positions etc.,
           * as defined by CSS2.)
           */
-         avail_width = table->avail_width *
-            MIN (DW_STYLE_GET_PERCENTAGE (widget->style->width), 1.0);
+         avail_width = table->avail_width
+            * MIN (DW_STYLE_PER_LENGTH_VAL (widget->style->width), 1.0);
       else
          avail_width = table->avail_width;
 
       if (avail_width < extremes.min_width)
          avail_width = extremes.min_width;
-      if (widget->style->width == DW_STYLE_UNDEF_LENGTH &&
+      if (widget->style->width == DW_STYLE_LENGTH_AUTO &&
           avail_width > extremes.max_width)
          avail_width = extremes.max_width;
 
@@ -305,6 +328,20 @@ static void Dw_table_size_request (DwWidget *widget,
       table->cum_height[row] = widget->style->border_spacing;
 
    for (row = 0; row < table->num_rows; row++) {
+      /* Calculate the baselines, for cells only with valign=baseline. */
+      table->baseline[row] = 0;
+      for (col = 0; col < table->num_cols; col++) {
+         n = col + row * table->num_cols;
+         if (CHILD_DEFINED (table, n)) {
+            child = table->children[n]->data.cell.widget;
+            if (child->style->valign == DW_STYLE_VALIGN_BASELINE) {
+               a_Dw_widget_size_request (child, &child_requisition);
+               if (child_requisition.ascent > table->baseline[row])
+                  table->baseline[row] = child_requisition.ascent;
+            }
+         }
+      }
+
       for (col = 0; col < table->num_cols; col++) {
          n = col + row * table->num_cols;
          if (CHILD_DEFINED (table, n)) {
@@ -321,8 +358,14 @@ static void Dw_table_size_request (DwWidget *widget,
             a_Dw_widget_set_descent (child, table->avail_descent);
 
             a_Dw_widget_size_request (child, &child_requisition);
-            child_height =
-               child_requisition.ascent + child_requisition.descent;
+            /* Cells with valign=baseline have a height which may
+             * be greater than the height returned by  size_request. */
+            if (child->style->valign == DW_STYLE_VALIGN_BASELINE)
+               child_height =
+                  table->baseline[row] + child_requisition.descent;
+            else
+               child_height =
+                  child_requisition.ascent + child_requisition.descent;
 
             row2 = row + table->children[n]->data.cell.rowspan;
 
@@ -369,8 +412,8 @@ static void Dw_table_get_extremes (DwWidget *widget,
          Dw_table_sub_create (table);
 
       DEBUG_MSG (DEBUG_EXTR_LEVEL + 3,
-                 "--> Dw_table_get_extremes for (%d x %d) table.\n",
-                 table->num_cols, table->num_rows);
+                 "--> [%p] Dw_table_get_extremes for (%d x %d) table.\n",
+                 table, table->num_cols, table->num_rows);
 
       Dw_table_sub_get_extremes (table->sub);
       diff = 2 * widget->style->border_spacing +
@@ -378,7 +421,7 @@ static void Dw_table_get_extremes (DwWidget *widget,
       extremes->min_width = table->sub->total_extremes.min_width + diff;
       extremes->max_width = table->sub->total_extremes.max_width + diff;
 
-      DEBUG_MSG (DEBUG_EXTR_LEVEL + 2, "<-- END\n");
+      DEBUG_MSG (DEBUG_EXTR_LEVEL + 2, "<-- [%p] END\n", table);
    } else {
       extremes->min_width = widget->style->border_spacing
          + p_Dw_style_box_diff_width (widget->style);
@@ -396,9 +439,10 @@ static void Dw_table_size_allocate (DwWidget *widget,
 {
    DwTable *table = DW_TABLE (widget);
    gint row, col, col2, n;
-   gint32 col_offset, cell_width, vdiff;
+   gint32 col_offset, cell_width, cell_height, height_diff, vdiff;
    DwWidget *child;
    DwAllocation child_allocation;
+   DwRequisition child_requisition;
 
    /*
     * todo: This implementation depends on that size_request has been
@@ -425,13 +469,45 @@ static void Dw_table_size_allocate (DwWidget *widget,
                cell_width += table->col_width[col2];
 
             child = table->children[n]->data.cell.widget;
+            cell_height =
+               table->cum_height[row + table->children[n]->data.cell.rowspan] -
+               table->cum_height[row] - widget->style->border_spacing;
             child_allocation.x = col_offset;
             child_allocation.y = vdiff + table->cum_height[row];
             child_allocation.width = cell_width;
-            child_allocation.ascent =
-               table->cum_height[row + table->children[n]->data.cell.rowspan] -
-               table->cum_height[row] - widget->style->border_spacing;
-            child_allocation.descent = 0;
+
+            /* Vertical alignment is done via allocation. */
+            a_Dw_widget_size_request (child, &child_requisition);
+            switch (child->style->valign) {
+            case DW_STYLE_VALIGN_TOP:
+               child_allocation.ascent = child_requisition.ascent;
+               child_allocation.descent =
+                  cell_height - child_requisition.ascent;
+               break;
+            case DW_STYLE_VALIGN_BOTTOM:
+               child_allocation.ascent =
+                  cell_height - child_requisition.descent;
+               child_allocation.descent = child_requisition.descent;
+               break;
+            case DW_STYLE_VALIGN_MIDDLE:
+               height_diff = cell_height - (child_requisition.ascent +
+                                            child_requisition.descent);
+               child_allocation.ascent =
+                  child_requisition.ascent + height_diff / 2;
+               child_allocation.descent =
+                  child_requisition.descent + (height_diff + 1) / 2;
+               /* The "+ 1" is to avoid rounding errors. */
+               break;
+            case DW_STYLE_VALIGN_BASELINE:
+               child_allocation.ascent = table->baseline[row];
+               child_allocation.descent =
+                  cell_height - table->baseline[row];
+               break;
+            default:
+               g_assert_not_reached();
+               break;
+            }
+
             a_Dw_widget_size_allocate (child, &child_allocation);
          }
 
@@ -504,18 +580,19 @@ static void Dw_table_draw (DwWidget *widget,
    gint i;
    gint32 offx, offy, width;
 
-   p_Dw_widget_draw_widget_box (widget, area);
+   p_Dw_widget_draw_widget_box (widget, area, FALSE);
 
    offx = p_Dw_style_box_offset_x (widget->style);
    offy = p_Dw_style_box_offset_y (widget->style);
-   width = widget->allocation.width - p_Dw_style_box_diff_width(widget->style);
+   width = DW_WIDGET_CONTENT_WIDTH(widget);
+
    for (i = 0; i < table->num_rows; i++)
       if (table->row_style[i])
          p_Dw_widget_draw_box (widget, table->row_style[i], area,
                                offx, offy + table->cum_height[i],
                                width,
                                table->cum_height[i + 1] - table->cum_height[i]
-                               - widget->style->border_spacing);
+                               - widget->style->border_spacing, FALSE);
 
    for (i = 0; i < table->num_children; i++)
       if (CHILD_DEFINED (table, i)) {
@@ -529,7 +606,7 @@ static void Dw_table_draw (DwWidget *widget,
                 Dw_widget_x_world_to_viewport (child, child->allocation.x) - 1,
                 Dw_widget_y_world_to_viewport (child, child->allocation.y) - 1,
                 child->allocation.width + 2,
-                child->allocation.ascent + child->allocation.descent + 2);
+                DW_WIDGET_HEIGHT(child) + 2);
 #endif
             a_Dw_widget_draw (child, &child_area, event);
          }
@@ -543,11 +620,189 @@ static void Dw_table_draw (DwWidget *widget,
                        Dw_widget_y_world_to_viewport (widget,
                                                       widget->allocation.y),
                        widget->allocation.width,
-                       widget->allocation.ascent + widget->allocation.descent);
+                       DW_WIDGET_HEIGHT(widget));
 #endif
 }
 
 
+/*
+ * Construct an iterator for a cell.
+ */
+static DwIterator *Dw_table_construct_iterator (DwTable *table,
+                                                gint cell_index)
+{
+   DwIterator *it;
+
+   it = Dw_table_iterator (DW_WIDGET (table), DW_CONTENT_ALL, FALSE);
+   ((DwIteratorInt*)it)->pos = cell_index;
+   it->content.type = DW_CONTENT_WIDGET;
+   it->content.data.widget = table->children[cell_index]->data.cell.widget;
+   return it;
+}
+
+/*
+ * Construct an iterator pointing to the end of the table.
+ */
+static DwIterator *Dw_table_construct_last_iterator (DwTable *table,
+                                                     int *char_pos)
+{
+   int n;
+   DwIterator *it = NULL;
+   DwTableChild *child = NULL;
+
+   *char_pos = 0;
+   n = table->num_children -1;
+   while (n >= 0 && ((child = table->children[n]) == NULL ||
+                     child->type != DW_TABLE_CELL))
+      n--;
+   if (n >= 0) {
+      /* Try to construct an iterator for the last child of the table. */
+      it = a_Dw_widget_iterator (child->data.cell.widget,
+                                 DW_CONTENT_ALL, TRUE);
+      if (it != NULL && a_Dw_iterator_prev (it)) {
+         if (it->content.type == DW_CONTENT_TEXT)
+            *char_pos = strlen (it->content.data.text);
+      } else {
+         a_Dw_iterator_free (it);
+         it = NULL;
+      }
+
+      if (it == NULL)
+         /* This widget has no content, construct an iterator for the table. */
+         it = Dw_table_construct_iterator (table, n);
+   }
+
+   return it;
+}
+
+/*
+ * Send event to selection.
+ */
+static gboolean Dw_table_send_selection_event (DwTable *table,
+                                               gint (*fn) (Selection*,
+                                                           DwIterator *,
+                                                           gint, gint,
+                                                           GdkEventButton*,
+                                                           gboolean
+                                                           within_content),
+                                               gint32 x, gint32 y,
+                                               GdkEventButton *event)
+{
+   int row, col, n, char_pos;
+   DwIterator *it;
+   DwTableChild *child;
+   gint32 offy, cellspacing, cum_width;
+
+   it = NULL;
+   char_pos = 0;
+   cellspacing = DW_WIDGET(table)->style->border_spacing;
+
+   /* Search row.
+    * Just a note: Improving the speed of this does not give very much,
+    * since events (in dw_widget.c) are always much slower. */
+   offy = p_Dw_style_box_offset_y (DW_WIDGET(table)->style);
+   row = 0;
+   while (row < table->num_rows &&
+          y > table->cum_height[row + 1] + offy - cellspacing)
+      row++;
+
+   if (row == table->num_rows)
+      /* Below the last row: We create an iterator of the last child,
+       * starting at the end (if possible). */
+      it = Dw_table_construct_last_iterator (table, &char_pos);
+   else {
+      if (y < table->cum_height[row] + offy) {
+         /* Before the row: Start at first column. */
+         col = 0;
+      } else {
+         /* Within the row: Search column.
+          * NOTE: col may be table->num_cols, this is intended, see below */
+         cum_width =
+            p_Dw_style_box_offset_x (DW_WIDGET(table)->style) +
+            cellspacing + table->col_width[0];
+         col = 0;
+         while (col < table->num_cols && x > cum_width) {
+            col++;
+            cum_width += table->col_width[col] + cellspacing;
+         }
+
+         /* Already right of the middle of the next column?
+          * (Actually, this feature is only needed for cases where widgets
+          * are added to a table, which do not process mouse events. This
+          * is never the case in dillo. But it works.) */
+         if (col < table->num_cols &&
+             x >= cum_width - table->col_width[col] / 2)
+            col++;
+      }
+
+      n = col + row * table->num_cols;
+      /* If col == table->num_cols, this may be happen: */
+      if (n == table->num_children)
+         it = Dw_table_construct_last_iterator (table, &char_pos);
+      else {
+         /* some corrections */
+         for (; n < table->num_children; n++) {
+            if ((child = table->children[n]) &&
+                child->type == DW_TABLE_CELL) {
+               it = Dw_table_construct_iterator (table, n);
+               break;
+            }
+         }
+      }
+   }
+
+   if (it) {
+      return fn (GTK_DW_VIEWPORT(DW_WIDGET(table)->viewport)->selection,
+                 it, char_pos, -1, event, FALSE);
+   } else
+      return FALSE;
+}
+
+
+/*
+ * Standard Dw function.
+ */
+static gboolean Dw_table_button_press (DwWidget *widget,
+                                       gint32 x,
+                                       gint32 y,
+                                       GdkEventButton *event)
+{
+   return Dw_table_send_selection_event (DW_TABLE (widget),
+                                         a_Selection_button_press, x, y,
+                                         event);
+}
+
+/*
+ * Standard Dw function.
+ */
+static gboolean Dw_table_button_release (DwWidget *widget,
+                                         gint32 x,
+                                         gint32 y,
+                                         GdkEventButton *event)
+{
+   return Dw_table_send_selection_event (DW_TABLE (widget),
+                                         a_Selection_button_release, x, y,
+                                         event);
+}
+
+
+/*
+ * Standard Dw function.
+ */
+static gboolean Dw_table_motion_notify (DwWidget *widget,
+                                        gint32 x,
+                                        gint32 y,
+                                        GdkEventMotion *event)
+{
+   if (event && (event->state & GDK_BUTTON1_MASK))
+      return Dw_table_send_selection_event (DW_TABLE (widget),
+                                            a_Selection_button_motion, x, y,
+                                            NULL);
+   else
+      return FALSE;
+}
+
+
 /*
  * Standard Dw function.
  */
@@ -584,70 +839,10 @@ static void Dw_table_forall (DwContainer *container,
 }
 
 
-/*
- * Standard DwContainer function.
- */
-static gint Dw_table_findtext (DwContainer *container,
-                               gpointer FP, gpointer KP,
-                               gchar *NewKey)
-{
-   DwWidget *widget;
-   DwTable *table = DW_TABLE (container);
-   FindData *F = *(FindData**)FP;
-
-   g_return_val_if_fail ((DW_WIDGET(table))->viewport != NULL && KP, 0);
-
-   if ( !F )
-      *(FindData **)FP = F = g_new0(FindData, 1);
-
-   if ( !F->Key || strcmp(F->Key->KeyStr, NewKey) ||
-        F->widget != DW_WIDGET (table) )
-      F->State = F_NewKey;
-
-   /* Let the FSM find the search string */
-   while ( 1 ) {
-      switch (F->State) {
-      case F_NewKey:
-         F->Key = KP;
-         F->widget = DW_WIDGET (table);
-         F->WordNum = 0;
-         F->State = F_Seek;
-         break;
-      case F_Seek:
-         for (   ; F->WordNum < table->num_children; F->WordNum++) {
-           if (CHILD_DEFINED (table, F->WordNum)){
-              widget = table->children[F->WordNum]->data.cell.widget;
-              if (DW_IS_CONTAINER (widget) &&
-                  a_Dw_container_findtext(DW_CONTAINER (widget),
-                                          (gpointer)&F->next,
-                                          (gpointer)F->Key,
-                                          NewKey)){
-                  F->State = F_Seek;
-                  return 1;
-              }
-           }
-         }
-         F->State = F_End;
-         break;
-      case F_GetPos:
-      case F_Found:
-      case F_End:
-         /* free memory */
-         g_free(F);
-         *(FindData **)FP = NULL;
-         return 0;
-      }
-   }
-
-   /* compiler happiness */
-   return 0;
-}
-
-
 /*
  * Make sure that there is enough space for table->children,
- * table->cum_height and table->row_style, and fill the rest of
- * table->children with NULL.
+ * table->cum_height etc., and fill the rest of table->children with
+ * NULL.
  */
 static void Dw_table_realloc_children (DwTable *table)
 {
@@ -657,6 +852,7 @@ static void Dw_table_realloc_children (DwTable *table)
    a_List_resize (table->cum_height, table->num_rows, table->cum_height_max);
    table->num_rows--;
    a_List_resize (table->row_style, table->num_rows, table->row_style_max);
+   a_List_resize (table->baseline, table->num_rows, table->baseline_max);
 
    old_num_children = table->num_children;
    table->num_children = table->num_cols * table->num_rows;
@@ -746,6 +942,9 @@ void a_Dw_table_add_cell (DwTable *table,
    table->cur_col += colspan;
 
    p_Dw_widget_set_parent (widget, DW_WIDGET (table));
+   if (table->row_style[table->cur_row] != NULL)
+      p_Dw_widget_set_bg_color
+         (widget, table->row_style[table->cur_row]->background_color);
    p_Dw_widget_queue_resize (widget, 0, TRUE);
 
    /* The cell structure has changed, so the subtables have to be rebuilt. */
@@ -784,6 +983,26 @@ void a_Dw_table_add_row (DwTable *table,
    table->cur_col = 0;
 }
 
+/*
+ * Returns the DwTableCell to be passed to a_Dw_table_cell_new().
+ */
+DwTableCell* a_Dw_table_get_cell_ref (DwTable *table)
+{
+   gint row, n;
+   DwWidget *child;
+
+   for (row = 0; row <= table->num_rows; row++) {
+      n = table->cur_col + row * table->num_cols;
+      if (CHILD_DEFINED (table, n)) {
+         child = table->children[n]->data.cell.widget;
+         if (DW_IS_TABLE_CELL (child))
+            return DW_TABLE_CELL (child);
+      }
+   }
+
+   return NULL;
+}
+
 
 /* ----------------------------------------------------------------------
  *
@@ -991,8 +1210,8 @@ static void Dw_table_sub_get_extremes (DwTableSub *sub)
    sub->fixed_width = FALSE;
    sub->percentage = 0;
 
-   DEBUG_MSG (DEBUG_EXTR_LEVEL + 2, "   subtable from %d to %d\n",
-              sub->start_col, sub->end_col);
+   DEBUG_MSG (DEBUG_EXTR_LEVEL + 2, "   [%p] subtable from %d to %d\n",
+              sub->table, sub->start_col, sub->end_col);
 
    /* 1. cells spanning the whole width */
    sub->span_extremes.min_width = 0;
@@ -1018,24 +1237,24 @@ static void Dw_table_sub_get_extremes (DwTableSub *sub)
                colspan = sub->table->children[start_n]->data.cell.colspan;
 
                /* Adjust width argument of the cell */
-               if (DW_STYLE_IS_LENGTH (child->style->width)) {
-                  col_width = DW_STYLE_GET_LENGTH (child->style->width,
-                                                   child->style->font);
+               if (DW_STYLE_IS_ABS_LENGTH (child->style->width)) {
+                  col_width =
+                     DW_STYLE_ABS_LENGTH_VAL (child->style->width);
                   if (col_width < child_extremes.min_width)
                      col_width = child_extremes.min_width;
                   child_extremes.min_width = col_width;
                   child_extremes.max_width = col_width;
                   sub->fixed_width = TRUE;
-                  DEBUG_MSG (DEBUG_EXTR_LEVEL,
-                             "         following adjusted by %d pixels:\n",
-                             col_width);
-               } else if (DW_STYLE_IS_PERCENTAGE (child->style->width)) {
-                  percentage = DW_STYLE_GET_PERCENTAGE (child->style->width);
+                  DEBUG_MSG (DEBUG_EXTR_LEVEL, "         [%p] following "
+                             "adjusted by %d pixels:\n",
+                             sub->table, col_width);
+               } else if (DW_STYLE_IS_PER_LENGTH (child->style->width)) {
+                  percentage = DW_STYLE_PER_LENGTH_VAL (child->style->width);
                   sub->percentage = MAX (sub->percentage, percentage);
                   sub->use_percentage = USE_PERCENTAGE_ALL;
                   DEBUG_MSG (DEBUG_EXTR_LEVEL,
-                             "         following adjusted by %g %%\n",
-                             100 * sub->percentage);
+                             "         [%p] following adjusted by %g %%\n",
+                             sub->table, 100 * sub->percentage);
                }
 
                if (num_cols != colspan) {
@@ -1046,9 +1265,9 @@ static void Dw_table_sub_get_extremes (DwTableSub *sub)
                }
 
                DEBUG_MSG (DEBUG_EXTR_LEVEL,
-                          "         row %d: cell extremes: %d / %d "
+                          "         [%p], row %d: cell extremes: %d / %d "
                           "[was multiplied by %d / %d]\n",
-                          row, child_extremes.min_width,
+                          sub->table, row, child_extremes.min_width,
                           child_extremes.max_width,
                           num_cols, colspan);
 
@@ -1058,22 +1277,23 @@ static void Dw_table_sub_get_extremes (DwTableSub *sub)
                   sub->span_extremes.max_width = child_extremes.max_width;
             } else {
                DEBUG_MSG (DEBUG_EXTR_LEVEL,
-                          "         row %d: omitted\n", row);
+                          "         [%p] row %d: omitted\n",
+                          sub->table, row);
             }
          }
       }
    }
 
-   DEBUG_MSG (DEBUG_EXTR_LEVEL + 1, "      spanning subs: %d / %d\n",
-              sub->span_extremes.min_width, sub->span_extremes.max_width);
+   DEBUG_MSG (DEBUG_EXTR_LEVEL + 1, "      [%p] spanning subs: %d / %d\n",
+              sub->table, sub->span_extremes.min_width,
+              sub->span_extremes.max_width);
 
    if (sub->table->sub == sub) {
       /* Adjust width argument of the table (fixed, percentages
          later in the code). */
-      widget = DW_WIDGET(sub->table);
-      if (DW_STYLE_IS_LENGTH (widget->style->width)) {
-         col_width = DW_STYLE_GET_LENGTH (widget->style->width,
-                                          widget->style->font);
+      widget = DW_WIDGET (sub->table);
+      if (DW_STYLE_IS_ABS_LENGTH (widget->style->width)) {
+         col_width = DW_STYLE_ABS_LENGTH_VAL (widget->style->width);
          if (col_width < sub->span_extremes.min_width)
             col_width = sub->span_extremes.min_width;
          sub->span_extremes.min_width = col_width;
@@ -1099,8 +1319,8 @@ static void Dw_table_sub_get_extremes (DwTableSub *sub)
                (i == 0) ? USE_PERCENTAGE_ALL : USE_PERCENTAGE_SOME;
       }
 
-   DEBUG_MSG (DEBUG_EXTR_LEVEL + 1, "      sum of subsubs: %d / %d\n",
-              sum_extremes.min_width, sum_extremes.max_width);
+   DEBUG_MSG (DEBUG_EXTR_LEVEL + 1, "      [%p] sum of subsubs: %d / %d\n",
+              sub->table, sum_extremes.min_width, sum_extremes.max_width);
 
    if (sub->fixed_width) {
       sum_extremes.max_width = sum_extremes.min_width;
@@ -1113,8 +1333,9 @@ static void Dw_table_sub_get_extremes (DwTableSub *sub)
                                         sub->span_extremes.max_width);
    sub->percentage = MAX (sub->percentage, sum_percentage);
 
-   DEBUG_MSG (DEBUG_EXTR_LEVEL + 1, "      final: %d / %d\n",
-              sub->total_extremes.min_width, sub->total_extremes.max_width);
+   DEBUG_MSG (DEBUG_EXTR_LEVEL + 1, "      [%p] final: %d / %d\n",
+              sub->table, sub->total_extremes.min_width,
+              sub->total_extremes.max_width);
 }
 
 /*
@@ -1167,7 +1388,10 @@ static void Dw_table_sub_adjust_col_widths (DwTableSub *sub,
               sub_extr_width = rest_w * cols_per_sub / rest_n;
               rest_w -= sub_extr_width;
               rest_n -= cols_per_sub;
-              EXTR_VALUE (sub_extremes[i]) = sub_extr_width;
+              if (max)
+                 sub_extremes[i].max_width = sub_extr_width;
+              else
+                 sub_extremes[i].min_width = sub_extr_width;
             }
          }
       } else {
@@ -1186,7 +1410,10 @@ static void Dw_table_sub_adjust_col_widths (DwTableSub *sub,
                   delta = rest_w * EXTR_VALUE (sub_extremes[i]) / rest_n;
                   rest_w -= delta;
                   rest_n -= EXTR_VALUE (sub_extremes[i]);
-                  EXTR_VALUE (sub_extremes[i]) += delta;
+                  if (max)
+                     sub_extremes[i].max_width += delta;
+                  else
+                     sub_extremes[i].min_width += delta;
                }
                DEBUG_MSG (DEBUG_WIDTH_LEVEL + 2, "%d\n",
                           EXTR_VALUE (sub_extremes[i]));
@@ -1232,7 +1459,7 @@ static void Dw_table_sub_calc_col_widths (DwTableSub *sub,
                  "(total = %d)\n",
                  sub->start_col, sub->end_col, width, total_width);
 
-      border_spacing = DW_WIDGET(sub->table)->style->border_spacing;
+      border_spacing = DW_WIDGET (sub->table)->style->border_spacing;
       diff =  (sub->num_subs - 1) * border_spacing;
       avail_width = width - diff;
 
@@ -1441,3 +1668,92 @@ static void Dw_table_sub_calc_col_widths (DwTableSub *sub,
       g_free (sub_extremes);
    }
 }
+
+
+static DwIterator *Dw_table_iterator (DwWidget *widget,
+                                      gint mask,
+                                      gboolean at_end)
+{
+   DwIteratorInt *it = g_new (DwIteratorInt, 1);
+   it->it.widget = widget;
+   it->it.mask = mask;
+   it->it.next = Dw_table_iterator_next;
+   it->it.prev = Dw_table_iterator_prev;
+   it->it.clone = p_Dw_iterator_clone_std_int;
+   it->it.compare = p_Dw_iterator_compare_std_int;
+   it->it.free = p_Dw_iterator_free_std;
+   it->it.highlight = p_Dw_iterator_highlight_std;
+   it->it.get_allocation = p_Dw_iterator_get_allocation_std_only_widgets;
+
+   if (at_end) {
+      it->it.content.type = DW_CONTENT_END;
+      it->pos = DW_TABLE(widget)->num_children;
+   } else {
+      it->it.content.type = DW_CONTENT_START;
+      it->pos = -1;
+   }
+
+   return (DwIterator*)it;
+}
+
+static gboolean Dw_table_iterator_next (DwIterator *it)
+{
+   DwTable *table = DW_TABLE (it->widget);
+   DwIteratorInt *ii = (DwIteratorInt*)it;
+
+   DEBUG_MSG (DEBUG_ITERATOR_LEVEL, "Dw_table_iterator_next: %d of %d\n",
+              ii->pos, table->num_children);
+
+   if (it->content.type == DW_CONTENT_END)
+      return FALSE;
+
+   /* tables only contain widgets: */
+   if ((it->mask & DW_CONTENT_WIDGET) == 0) {
+      it->content.type = DW_CONTENT_END;
+      return FALSE;
+   }
+
+   do {
+      ii->pos++;
+      if (ii->pos >= table->num_children) {
+         it->content.type = DW_CONTENT_END;
+         return FALSE;
+      }
+   } while (table->children[ii->pos] == NULL ||
+            table->children[ii->pos]->type != DW_TABLE_CELL);
+
+   it->content.type = DW_CONTENT_WIDGET;
+   it->content.data.widget = table->children[ii->pos]->data.cell.widget;
+   return TRUE;
+}
+
+static gboolean Dw_table_iterator_prev (DwIterator *it)
+{
+   DwTable *table = DW_TABLE (it->widget);
+   DwIteratorInt *ii = (DwIteratorInt*)it;
+
+   if (it->content.type == DW_CONTENT_START)
+      return FALSE;
+
+   DEBUG_MSG (DEBUG_ITERATOR_LEVEL, "Dw_table_iterator_prev: %d of %d\n",
+              ii->pos, table->num_children);
+
+   /* tables only contain widgets: */
+   if ((it->mask & DW_CONTENT_WIDGET) == 0) {
+      it->content.type = DW_CONTENT_START;
+      return FALSE;
+   }
+
+   do {
+      ii->pos--;
+      if (ii->pos < 0) {
+         it->content.type = DW_CONTENT_START;
+         return FALSE;
+      }
+   } while (table->children[ii->pos] == NULL ||
+            table->children[ii->pos]->type != DW_TABLE_CELL);
+
+   it->content.type = DW_CONTENT_WIDGET;
+   it->content.data.widget = table->children[ii->pos]->data.cell.widget;
+   return TRUE;
+}
