@@ -6,8 +6,12 @@
 GtkWidget *a_Menu_mainbar_new (BrowserWindow *bw, gint tiny);
 GtkWidget *a_Menu_popup_op_new (BrowserWindow *bw);
 GtkWidget *a_Menu_popup_ol_new (BrowserWindow *bw);
+GtkWidget *a_Menu_popup_oi_new (BrowserWindow *bw);
+GtkWidget *a_Menu_popup_ob_new (BrowserWindow *bw);
 GtkWidget *a_Menu_popup_history_new(BrowserWindow *bw, gint direction);
-void a_Menu_popup_set_url(BrowserWindow *bw, DilloUrl *url);
+void a_Menu_popup_ol_show_oi(BrowserWindow *bw, gboolean show);
+void a_Menu_popup_set_url(BrowserWindow *bw, const DilloUrl *url);
+void a_Menu_popup_set_url2(BrowserWindow *bw, const DilloUrl *url);
 DilloUrl *a_Menu_popup_get_url(BrowserWindow *bw);
 
 void a_Menu_pagemarks_new (BrowserWindow *bw);
