@@ -24,10 +24,10 @@
 #include <stdlib.h>
 #include <string.h>       /* for strchr */
 #include <fcntl.h>
+#include <unistd.h>
 #include <locale.h>       /* for setlocale */
 #include "prefs.h"
 #include "colors.h"
-#include "dillo.h"
 #include "misc.h"
 
 /* symbol array */
@@ -37,6 +37,7 @@ static const struct {
 } symbols[] = {
    { "geometry", DRC_TOKEN_GEOMETRY },
    { "http_proxy", DRC_TOKEN_PROXY },
+   { "http_proxyuser", DRC_TOKEN_PROXYUSER },
    { "no_proxy", DRC_TOKEN_NOPROXY },
    { "link_color", DRC_TOKEN_LINK_COLOR },
    { "visited_color", DRC_TOKEN_VISITED_COLOR, },
@@ -47,7 +48,7 @@ static const struct {
    { "text_color", DRC_TOKEN_TEXT_COLOR },
    { "use_oblique", DRC_TOKEN_USE_OBLIQUE },
    { "home", DRC_TOKEN_HOME },
-   { "show_alt", DRC_TOKEN_SHOW_ALT },
+   { "show_tooltip", DRC_TOKEN_SHOW_TOOLTIP },
    { "panel_size", DRC_TOKEN_PANEL_SIZE },
    { "small_icons", DRC_TOKEN_SMALL_ICONS },
    { "limit_text_width", DRC_TOKEN_LIMIT_TEXT_WIDTH },
@@ -59,15 +60,21 @@ static const struct {
    { "show_reload", DRC_TOKEN_SHOW_RELOAD },
    { "show_save", DRC_TOKEN_SHOW_SAVE },
    { "show_stop", DRC_TOKEN_SHOW_STOP },
+   { "show_bookmarks", DRC_TOKEN_SHOW_BOOKMARKS },
    { "show_menubar", DRC_TOKEN_SHOW_MENUBAR },
    { "show_clear_url", DRC_TOKEN_SHOW_CLEAR_URL },
    { "show_url", DRC_TOKEN_SHOW_URL },
+   { "show_search", DRC_TOKEN_SHOW_SEARCH },
    { "show_progress_box", DRC_TOKEN_SHOW_PROGRESS_BOX },
+   { "fullwindow_start", DRC_TOKEN_FULLWINDOW_START },
    { "transient_dialogs", DRC_TOKEN_TRANSIENT_DIALOGS },
    { "vw_fontname", DRC_TOKEN_VW_FONT },
    { "fw_fontname", DRC_TOKEN_FW_FONT },
    { "generate_submit", DRC_TOKEN_GENERATE_SUBMIT },
-   { "enterpress_forces_submit", DRC_TOKEN_ENTERPRESS_FORCES_SUBMIT }
+   { "enterpress_forces_submit", DRC_TOKEN_ENTERPRESS_FORCES_SUBMIT },
+   { "search_url", DRC_TOKEN_SEARCH_URL },
+   { "show_msg", DRC_TOKEN_SHOW_MSG },
+   { "show_extra_warnings", DRC_TOKEN_SHOW_EXTRA_WARNINGS }
 };
 
 static const guint n_symbols = sizeof (symbols) / sizeof (symbols[0]);
@@ -79,6 +86,7 @@ static const guint n_symbols = sizeof (symbols) / sizeof (symbols[0]);
  */
 static guint Prefs_parser(GScanner *scanner)
 {
+   gint st;
    guint symbol;
 
    /* expect a valid symbol */
@@ -102,20 +110,17 @@ static guint Prefs_parser(GScanner *scanner)
 
    /* assign value and exit successfully */
    switch (symbol) {
-   case DRC_TOKEN_GEOMETRY: {
-      gchar *ptr, *str = scanner->value.v_string;
-      gint width, height;
-
-      if ( (ptr = strchr(str, 'x')) && (width = strtol(str,NULL,10)) &&
-           (height = strtol(++ptr,NULL,10)) ){
-         prefs.width = width;
-         prefs.height = height;
-      }
-     }
-     break;
+   case DRC_TOKEN_GEOMETRY:
+      a_Misc_parse_geometry(scanner->value.v_string, &prefs.xpos, &prefs.ypos,
+                            &prefs.width, &prefs.height);
+      break;
    case DRC_TOKEN_PROXY:
       a_Url_free(prefs.http_proxy);
-      prefs.http_proxy = a_Url_new(scanner->value.v_string, NULL, 0, 0);
+      prefs.http_proxy = a_Url_new(scanner->value.v_string, NULL, 0, 0, 0);
+      break;
+   case DRC_TOKEN_PROXYUSER:
+      g_free(prefs.http_proxyuser);
+      prefs.http_proxyuser = g_strdup(scanner->value.v_string);
       break;
    case DRC_TOKEN_NOPROXY:
       g_free(prefs.no_proxy);
@@ -124,19 +129,19 @@ static guint Prefs_parser(GScanner *scanner)
       break;
    case DRC_TOKEN_LINK_COLOR:
       prefs.link_color = a_Color_parse(scanner->value.v_string,
-                                       prefs.link_color);
+                                       prefs.link_color, &st);
       break;
    case DRC_TOKEN_VISITED_COLOR:
       prefs.visited_color = a_Color_parse(scanner->value.v_string,
-                                          prefs.visited_color);
+                                          prefs.visited_color, &st);
       break;
    case DRC_TOKEN_TEXT_COLOR:
       prefs.text_color = a_Color_parse(scanner->value.v_string,
-                                       prefs.text_color);
+                                       prefs.text_color, &st);
       break;
    case DRC_TOKEN_BG_COLOR:
       prefs.bg_color = a_Color_parse(scanner->value.v_string,
-                                     prefs.bg_color);
+                                     prefs.bg_color, &st);
       break;
    case DRC_TOKEN_ALLOW_WHITE_BG:
       prefs.allow_white_bg = (strcmp(scanner->value.v_string, "YES") == 0);
@@ -164,10 +169,10 @@ static guint Prefs_parser(GScanner *scanner)
       break;
    case DRC_TOKEN_HOME:
       a_Url_free(prefs.home);
-      prefs.home = a_Url_new(scanner->value.v_string, NULL, 0, 0);
+      prefs.home = a_Url_new(scanner->value.v_string, NULL, 0, 0, 0);
       break;
-   case DRC_TOKEN_SHOW_ALT:
-      prefs.show_alt = (strcmp(scanner->value.v_string, "YES") == 0);
+   case DRC_TOKEN_SHOW_TOOLTIP:
+      prefs.show_tooltip = (strcmp(scanner->value.v_string, "YES") == 0);
       break;
    case DRC_TOKEN_FONT_FACTOR:
       prefs.font_factor = strtod(scanner->value.v_string, NULL);
@@ -196,6 +201,9 @@ static guint Prefs_parser(GScanner *scanner)
    case DRC_TOKEN_SHOW_STOP:
       prefs.show_stop = (strcmp(scanner->value.v_string, "YES") == 0);
       break;
+   case DRC_TOKEN_SHOW_BOOKMARKS:
+      prefs.show_bookmarks = (strcmp(scanner->value.v_string, "YES") == 0);
+      break;
    case DRC_TOKEN_SHOW_MENUBAR:
       prefs.show_menubar = (strcmp(scanner->value.v_string, "YES") == 0);
       break;
@@ -205,9 +213,15 @@ static guint Prefs_parser(GScanner *scanner)
    case DRC_TOKEN_SHOW_URL:
       prefs.show_url = (strcmp(scanner->value.v_string, "YES") == 0);
       break;
+   case DRC_TOKEN_SHOW_SEARCH:
+      prefs.show_search = (strcmp(scanner->value.v_string, "YES") == 0);
+      break;
    case DRC_TOKEN_SHOW_PROGRESS_BOX:
       prefs.show_progress_box = (strcmp(scanner->value.v_string, "YES") == 0);
       break;
+   case DRC_TOKEN_FULLWINDOW_START:
+      prefs.fullwindow_start = (strcmp(scanner->value.v_string, "YES") == 0);
+      break;
    case DRC_TOKEN_TRANSIENT_DIALOGS:
       prefs.transient_dialogs = (strcmp(scanner->value.v_string, "YES") == 0);
       break;
@@ -226,6 +240,18 @@ static guint Prefs_parser(GScanner *scanner)
       prefs.enterpress_forces_submit =
          (strcmp(scanner->value.v_string, "YES") == 0);
       break;
+   case DRC_TOKEN_SEARCH_URL:
+      g_free(prefs.search_url);
+      prefs.search_url = g_strdup(scanner->value.v_string);
+      break;
+   case DRC_TOKEN_SHOW_MSG:
+      prefs.show_msg =
+         (strcmp(scanner->value.v_string, "YES") == 0);
+      break;
+   case DRC_TOKEN_SHOW_EXTRA_WARNINGS:
+      prefs.show_extra_warnings =
+         (strcmp(scanner->value.v_string, "YES") == 0);
+      break;
    default:
       break;   /* Not reached */
    }
@@ -243,9 +269,10 @@ static gint Prefs_load(void)
    file = a_Misc_prepend_user_home(".dillo/dillorc");
    fd = open(file, O_RDONLY);
    g_free(file);
-   if (fd < 0 && (fd = open("/etc/dillorc", O_RDONLY)) < 0)
-      if ((fd = open("/usr/local/etc/dillorc", O_RDONLY)) < 0)
-         return FILE_NOT_FOUND;
+   if (fd < 0 && (fd = open(DILLORC_SYS, O_RDONLY)) < 0)
+      return FILE_NOT_FOUND;
+
+   fcntl(fd, F_SETFD, FD_CLOEXEC | fcntl(fd, F_GETFD));
 
    scanner = g_scanner_new(NULL);
 
@@ -253,7 +280,7 @@ static gint Prefs_load(void)
    /* Specifies the chars which can be used in identifiers */
    scanner->config->cset_identifier_nth = (
       G_CSET_a_2_z
-      "~-_:&%#/.0123456789"
+      "~+-_:&%#/.0123456789"
       G_CSET_A_2_Z
       G_CSET_LATINS   /*??? I don't know if we need these two */
       G_CSET_LATINC   /*??? */
@@ -312,6 +339,7 @@ static gint Prefs_load(void)
 
    /* finish parsing */
    g_scanner_destroy(scanner);
+   close(fd);
    return PARSE_OK;
 }
 
@@ -321,7 +349,10 @@ void a_Prefs_init(void)
 
    prefs.width = D_GEOMETRY_DEFAULT_WIDTH;
    prefs.height = D_GEOMETRY_DEFAULT_HEIGHT;
+   prefs.xpos = D_GEOMETRY_DEFAULT_XPOS;
+   prefs.ypos = D_GEOMETRY_DEFAULT_YPOS;
    prefs.http_proxy = NULL;
+   prefs.http_proxyuser = NULL;
    prefs.no_proxy = NULL;
    prefs.no_proxy_vec = NULL;
    prefs.link_color = DW_COLOR_DEFAULT_BLUE;
@@ -329,11 +360,11 @@ void a_Prefs_init(void)
    prefs.bg_color = DW_COLOR_DEFAULT_BGND;
    prefs.text_color = DW_COLOR_DEFAULT_BLACK;
    prefs.use_oblique = FALSE;
-   prefs.home = a_Url_new(DILLO_HOME, NULL, 0, 0);
+   prefs.home = a_Url_new(DILLO_HOME, NULL, 0, 0, 0);
    prefs.allow_white_bg = TRUE;
    prefs.force_my_colors = FALSE;
    prefs.force_visited_color = FALSE;
-   prefs.show_alt = FALSE;
+   prefs.show_tooltip = FALSE;
    prefs.panel_size = 1;
    prefs.small_icons = FALSE;
    prefs.limit_text_width = FALSE;
@@ -345,15 +376,21 @@ void a_Prefs_init(void)
    prefs.show_reload=TRUE;
    prefs.show_save=TRUE;
    prefs.show_stop=TRUE;
+   prefs.show_bookmarks=TRUE;
    prefs.show_menubar=TRUE;
    prefs.show_clear_url=TRUE;
    prefs.show_url=TRUE;
+   prefs.show_search=TRUE;
    prefs.show_progress_box=TRUE;
+   prefs.fullwindow_start=FALSE;
    prefs.transient_dialogs=FALSE;
    prefs.vw_fontname = g_strdup("helvetica");
    prefs.fw_fontname = g_strdup("courier");
    prefs.generate_submit = FALSE;
    prefs.enterpress_forces_submit = FALSE;
+   prefs.search_url = g_strdup("http://www.google.com/search?q=%s");
+   prefs.show_msg = TRUE;
+   prefs.show_extra_warnings = FALSE;
 
    /* this locale stuff is to avoid parsing problems with float numbers */
    old_locale = g_strdup (setlocale (LC_NUMERIC, NULL));
@@ -371,6 +408,7 @@ void a_Prefs_init(void)
  */
 void a_Prefs_freeall(void)
 {
+   g_free(prefs.http_proxyuser);
    g_free(prefs.no_proxy);
    if (prefs.no_proxy_vec)
       g_strfreev(prefs.no_proxy_vec);
@@ -378,4 +416,5 @@ void a_Prefs_freeall(void)
    g_free(prefs.fw_fontname);
    g_free(prefs.vw_fontname);
    a_Url_free(prefs.home);
+   g_free(prefs.search_url);
 }
