@@ -7,8 +7,11 @@
 extern "C" {
 #endif /* __cplusplus */
 
+#define DILLO_HOME "http://www.dillo.org/"
 #define D_GEOMETRY_DEFAULT_WIDTH   640
 #define D_GEOMETRY_DEFAULT_HEIGHT  550
+#define D_GEOMETRY_DEFAULT_XPOS  -9999
+#define D_GEOMETRY_DEFAULT_YPOS  -9999
 
 #define DW_COLOR_DEFAULT_GREY   0xd6d6d6
 #define DW_COLOR_DEFAULT_BLACK  0x000000
@@ -27,6 +30,7 @@ typedef enum {
    DRC_TOKEN_FIRST = G_TOKEN_LAST,
    DRC_TOKEN_GEOMETRY,
    DRC_TOKEN_PROXY,
+   DRC_TOKEN_PROXYUSER,
    DRC_TOKEN_NOPROXY,
    DRC_TOKEN_LINK_COLOR,
    DRC_TOKEN_VISITED_COLOR,
@@ -40,7 +44,7 @@ typedef enum {
    DRC_TOKEN_PANEL_SIZE,
    DRC_TOKEN_SMALL_ICONS,
    DRC_TOKEN_FONT_FACTOR,
-   DRC_TOKEN_SHOW_ALT,
+   DRC_TOKEN_SHOW_TOOLTIP,
    DRC_TOKEN_LIMIT_TEXT_WIDTH,
    DRC_TOKEN_USE_DICACHE,
    DRC_TOKEN_SHOW_BACK,
@@ -49,15 +53,21 @@ typedef enum {
    DRC_TOKEN_SHOW_RELOAD,
    DRC_TOKEN_SHOW_SAVE,
    DRC_TOKEN_SHOW_STOP,
+   DRC_TOKEN_SHOW_BOOKMARKS,
    DRC_TOKEN_SHOW_MENUBAR,
    DRC_TOKEN_SHOW_CLEAR_URL,
    DRC_TOKEN_SHOW_URL,
+   DRC_TOKEN_SHOW_SEARCH,
    DRC_TOKEN_SHOW_PROGRESS_BOX,
+   DRC_TOKEN_FULLWINDOW_START,
    DRC_TOKEN_TRANSIENT_DIALOGS,
    DRC_TOKEN_FW_FONT,
    DRC_TOKEN_VW_FONT,
    DRC_TOKEN_GENERATE_SUBMIT,
    DRC_TOKEN_ENTERPRESS_FORCES_SUBMIT,
+   DRC_TOKEN_SEARCH_URL,
+   DRC_TOKEN_SHOW_MSG,
+   DRC_TOKEN_SHOW_EXTRA_WARNINGS,
 
    DRC_TOKEN_LAST
 } Dillo_Rc_TokenType;
@@ -67,7 +77,10 @@ typedef struct _DilloPrefs DilloPrefs;
 struct _DilloPrefs {
    gint width;
    gint height;
+   gint xpos;
+   gint ypos;
    DilloUrl *http_proxy;
+   gchar *http_proxyuser;
    gchar *no_proxy;
    gchar **no_proxy_vec;
    DilloUrl *home;
@@ -79,7 +92,7 @@ struct _DilloPrefs {
    gboolean use_oblique;
    gboolean force_my_colors;
    gboolean force_visited_color;
-   gboolean show_alt;
+   gboolean show_tooltip;
    gint panel_size;
    gboolean small_icons;
    gboolean limit_text_width;
@@ -91,15 +104,21 @@ struct _DilloPrefs {
    gboolean show_reload;
    gboolean show_save;
    gboolean show_stop;
+   gboolean show_bookmarks;
    gboolean show_menubar;
    gboolean show_clear_url;
    gboolean show_url;
+   gboolean show_search;
    gboolean show_progress_box;
+   gboolean fullwindow_start;
    gboolean transient_dialogs;
    gchar *vw_fontname;
    gchar *fw_fontname;
    gboolean generate_submit;
    gboolean enterpress_forces_submit;
+   gchar *search_url;
+   gboolean show_msg;
+   gboolean show_extra_warnings;
 };
 
 /* Global Data */
