@@ -1,7 +1,7 @@
 /*
  * File: url.c
  *
- * Copyright (C) 2001 Jorge Arellano Cid <jcid@inf.utfsm.cl>
+ * Copyright (C) 2001 Jorge Arellano Cid <jcid@dillo.org>
  *               2001 Livio Baldini Soares <livio@linux.ime.usp.br>
  *
  * This program is free software; you can redistribute it and/or modify
@@ -120,7 +120,7 @@ static DilloUrl *Url_object_new(const gchar *uri_str)
       s = ++p;
    }
    /* p = strpbrk(s, "/"); */
-   if (p && p[0] == '/' && p[1] == '/') {          /* authority */
+   if (p == s && p[0] == '/' && p[1] == '/') {     /* authority */
       s = p + 2;
       p = strpbrk(s, "/?#");
       if (p) {
@@ -186,7 +186,7 @@ static GString *Url_resolve_relative(const gchar *RelStr,
    gchar *p, *s, *e;
    gint i;
    GString *SolvedUrl, *Path;
-   DilloUrl *RelUrl = NULL, *BaseUrl = NULL;
+   DilloUrl *RelUrl, *BaseUrl = NULL;
 
    /* parse relative URL */
    RelUrl = Url_object_new(RelStr);
@@ -330,37 +330,46 @@ done:
  *  Return NULL if URL is badly formed.
  */
 DilloUrl* a_Url_new(const gchar *url_str, const gchar *base_url,
-                    gint flags, gint pos)
+                    gint flags, gint32 posx, gint32 posy)
 {
    DilloUrl *url;
-   gchar *urlstring, *p, *new_url = NULL;
+   gchar *urlstring, *p, *s;
    GString *SolvedUrl;
+   gint n_ic;
 
    g_return_val_if_fail (url_str != NULL, NULL);
 
-   /* let's use a heuristic to set http: as default */
+   /* denounce illegal characters (0x00-0x1F, 0x7F and space) */
    urlstring = (gchar *)url_str;
+   for (p = s = urlstring; *s; s++)
+      if (*s > 0x1F && *s != 0x7F && *s != ' ')
+         p++;
+   /* number of illegal characters */
+   n_ic = s - p;
+
+   /* let's use a heuristic to set http: as default */
    if (!base_url) {
       base_url = "http:";
       if (urlstring[0] != '/') {
          p = strpbrk(urlstring, "/#?:");
          if (!p || *p != ':')
-            urlstring = new_url = g_strconcat("//", urlstring, NULL);
+            urlstring = g_strconcat("//", urlstring, NULL);
       } else if (urlstring[1] != '/')
-         urlstring = new_url = g_strconcat("/", urlstring, NULL);
+         urlstring = g_strconcat("/", urlstring, NULL);
    }
 
    /* Resolve the URL */
    SolvedUrl = Url_resolve_relative(urlstring, NULL, base_url);
    DEBUG_MSG(2, "SolvedUrl = %s\n", SolvedUrl->str);
-   g_free(new_url);
    g_return_val_if_fail (SolvedUrl != NULL, NULL);
 
    /* Fill url data */
    url = Url_object_new(SolvedUrl->str);
    url->url_string = SolvedUrl;
    url->flags = flags;
-   url->scrolling_position = pos;
+   url->scrolling_position_x = posx;
+   url->scrolling_position_y = posy;
+   url->illegal_chars = n_ic;
 
    return url;
 }
@@ -376,13 +385,14 @@ DilloUrl* a_Url_dup(const DilloUrl *ori)
    url = Url_object_new(URL_STR_(ori));
    g_return_val_if_fail (url != NULL, NULL);
 
-   url->url_string         = g_string_new(URL_STR(ori));
-   url->port               = ori->port;
-   url->flags              = ori->flags;
-   url->data               = g_strdup(ori->data);
-   url->alt                = g_strdup(ori->alt);
-   url->ismap_url_len      = ori->ismap_url_len;
-   url->scrolling_position = ori->scrolling_position;
+   url->url_string           = g_string_new(URL_STR(ori));
+   url->port                 = ori->port;
+   url->flags                = ori->flags;
+   url->data                 = g_strdup(ori->data);
+   url->alt                  = g_strdup(ori->alt);
+   url->ismap_url_len        = ori->ismap_url_len;
+   url->scrolling_position_x = ori->scrolling_position_x;
+   url->scrolling_position_y = ori->scrolling_position_y;
 
    return url;
 }
@@ -436,8 +446,7 @@ void a_Url_set_data(DilloUrl *u, gchar *data)
 void a_Url_set_alt(DilloUrl *u, const gchar *alt)
 {
    if (u) {
-      if (u->alt)
-         g_free((gchar *)u->alt);
+      g_free((gchar *)u->alt);
       u->alt = g_strdup(alt);
    }
 }
@@ -445,10 +454,12 @@ void a_Url_set_alt(DilloUrl *u, const gchar *alt)
 /*
  * Set DilloUrl scrolling position
  */
-void a_Url_set_pos(DilloUrl *u, gint pos)
+void a_Url_set_pos(DilloUrl *u, gint32 posx, gint32 posy)
 {
-   if (u)
-      u->scrolling_position = pos;
+   if (u) {
+      u->scrolling_position_x = posx;
+      u->scrolling_position_y = posy;
+   }
 }
 
 /*
@@ -518,3 +529,30 @@ gchar *a_Url_parse_hex_path(const DilloUrl *u)
    new_uri = g_realloc(new_uri, sizeof(gchar) * (dest - new_uri));
    return new_uri;
 }
+
+/*
+ * RFC-2396 suggests this stripping when "importing" URLs from other media.
+ * Strip: "URL:", enclosing < >, and embedded whitespace.
+ * (We also strip illegal chars: 00-1F and 7F)
+ */
+gchar *a_Url_string_strip_delimiters(gchar *str)
+{
+   gchar *p, *new_str, *text;
+
+   new_str = text = g_strdup(str);
+
+   if (new_str) {
+      if (strncmp(new_str, "URL:", 4) == 0)
+         text += 4;
+      if (*text == '<')
+         text++;
+
+      for (p = new_str; *text; text++)
+         if (*text > 0x1F && *text != 0x7F && *text != ' ')
+            *p++ = *text;
+      if (p > new_str && p[-1] == '>')
+         --p;
+      *p = 0;
+   }
+   return new_str;
+}
